import warnings


def import_pytorch_lightning():
    # filter deprecation warnings
    warnings.filterwarnings("ignore", category=DeprecationWarning)
    import pytorch_lightning as pl

    warnings.filterwarnings("default", category=DeprecationWarning)
    return pl



import itertools
import unittest
from fractions import Fraction

import numpy as np

import neurosym as ns

from .bigram_test import fam, fam_with_ordering, fam_with_ordering_231, fam_with_vars

arith_dist = ns.TreeDistribution(
    1,
    {
        ((0, 0),): [(1, np.log(1 / 4)), (2, np.log(3 / 4))],
        ((1, 0),): [(1, np.log(1 / 4)), (2, np.log(3 / 4))],
        ((1, 1),): [(1, np.log(1 / 8)), (2, np.log(7 / 8))],
    },
    [("root", 1), ("+", 2), ("1", 0)],
    ns.NoopPreorderMask,
    ns.DefaultNodeOrdering,
)


def enumerated(*args, **kwargs):
    return [
        (ns.render_s_expression(program), likelihood)
        for program, likelihood in itertools.islice(
            ns.enumerate_tree_dist(*args, **kwargs), 100
        )
    ]


class TreeDistributionTest(unittest.TestCase):
    def enumerate_dsl(self, family, dist, min_likelihood=-6):
        result = list(family.enumerate(dist, min_likelihood=min_likelihood))
        result = sorted(result, key=lambda x: x[1], reverse=True)
        result = [
            (
                ns.render_s_expression(prog),
                Fraction(*np.exp(likelihood).as_integer_ratio()).limit_denominator(),
            )
            for prog, likelihood in result
        ]
        print(result)
        return set(result)

    def test_uniqueness(self):
        programs = enumerated(arith_dist)
        assert len(programs) == len(set(programs))

    def test_enumeration(self):
        out = dict(enumerated(arith_dist))
        p_one = 3 / 4
        p_add = 1 / 4
        self.assertAlmostEqual(out["(1)"], np.log(p_one))
        self.assertAlmostEqual(out["(+ (1) (1))"], np.log(p_add * 3 / 4 * 7 / 8))
        self.assertAlmostEqual(
            out["(+ (1) (+ (1) (1)))"],
            np.log(p_add * (3 / 4) * (1 / 8 * 3 / 4 * 7 / 8)),
        )
        self.assertAlmostEqual(
            out["(+ (+ (1) (1)) (1))"],
            np.log(p_add * (1 / 4 * 3 / 4 * 7 / 8) * (7 / 8)),
        )

    def test_enumeration_from_dsl_uniform(self):
        result = self.enumerate_dsl(fam, fam.uniform())

        self.assertEqual(
            result,
            {
                ("(1)", Fraction(1, 3)),
                ("(2)", Fraction(1, 3)),
                ("(+ (1) (1))", Fraction(1, 27)),
                ("(+ (2) (1))", Fraction(1, 27)),
                ("(+ (1) (2))", Fraction(1, 27)),
                ("(+ (2) (2))", Fraction(1, 27)),
                ("(+ (1) (+ (1) (1)))", Fraction(1, 243)),
                ("(+ (2) (+ (1) (1)))", Fraction(1, 243)),
                ("(+ (1) (+ (2) (1)))", Fraction(1, 243)),
                ("(+ (2) (+ (2) (1)))", Fraction(1, 243)),
                ("(+ (1) (+ (1) (2)))", Fraction(1, 243)),
                ("(+ (2) (+ (1) (2)))", Fraction(1, 243)),
                ("(+ (1) (+ (2) (2)))", Fraction(1, 243)),
                ("(+ (2) (+ (2) (2)))", Fraction(1, 243)),
                ("(+ (+ (1) (1)) (1))", Fraction(1, 243)),
                ("(+ (+ (2) (1)) (1))", Fraction(1, 243)),
                ("(+ (+ (1) (2)) (1))", Fraction(1, 243)),
                ("(+ (+ (2) (2)) (1))", Fraction(1, 243)),
                ("(+ (+ (1) (1)) (2))", Fraction(1, 243)),
                ("(+ (+ (2) (1)) (2))", Fraction(1, 243)),
                ("(+ (+ (1) (2)) (2))", Fraction(1, 243)),
                ("(+ (+ (2) (2)) (2))", Fraction(1, 243)),
            },
        )

    def test_enumeration_from_dsl_fitted_to_single(self):
        result = self.enumerate_dsl(
            fam,
            fam.counts_to_distribution(
                fam.count_programs([[ns.parse_s_expression("(+ (1) (2))")]])
            )[0],
        )
        self.assertEqual(
            result,
            {("(+ (1) (2))", Fraction(1, 1))},
        )

    def test_enumeration_from_dsl_fitted_to_two(self):
        result = self.enumerate_dsl(
            fam,
            fam.counts_to_distribution(
                fam.count_programs(
                    [[ns.parse_s_expression(x) for x in ("(+ (1) (2))", "(+ (2) (1))")]]
                )
            )[0],
        )
        self.assertEqual(
            result,
            {
                ("(+ (1) (1))", Fraction(1, 4)),
                ("(+ (1) (2))", Fraction(1, 4)),
                ("(+ (2) (1))", Fraction(1, 4)),
                ("(+ (2) (2))", Fraction(1, 4)),
            },
        )

    def test_enumeration_from_dsl_fitted_to_nested(self):
        result = self.enumerate_dsl(
            fam,
            fam.counts_to_distribution(
                fam.count_programs(
                    [
                        [
                            ns.parse_s_expression(x)
                            for x in ("(+ (1) (2))", "(+ (1) (+ (1) (2)))")
                        ]
                    ]
                )
            )[0],
        )
        self.assertEqual(
            result,
            {
                ("(+ (1) (2))", Fraction(2, 3)),
                ("(+ (1) (+ (1) (2)))", Fraction(2, 9)),
                ("(+ (1) (+ (1) (+ (1) (2))))", Fraction(2, 27)),
                ("(+ (1) (+ (1) (+ (1) (+ (1) (2)))))", Fraction(2, 81)),
                ("(+ (1) (+ (1) (+ (1) (+ (1) (+ (1) (2))))))", Fraction(2, 243)),
                (
                    "(+ (1) (+ (1) (+ (1) (+ (1) (+ (1) (+ (1) (2)))))))",
                    Fraction(2, 729),
                ),
            },
        )

    def test_enumeration_from_dsl_fitted_unbalanced(self):
        result = self.enumerate_dsl(
            fam,
            fam.counts_to_distribution(
                fam.count_programs(
                    [
                        [
                            ns.parse_s_expression(x)
                            for x in (
                                "(+ (1) (2))",
                                "(+ (1) (2))",
                                "(+ (1) (+ (1) (2)))",
                            )
                        ]
                    ]
                )
            )[0],
        )
        self.assertEqual(
            result,
            {
                ("(+ (1) (2))", Fraction(3, 4)),
                ("(+ (1) (+ (1) (2)))", Fraction(3, 16)),
                ("(+ (1) (+ (1) (+ (1) (2))))", Fraction(3, 64)),
                ("(+ (1) (+ (1) (+ (1) (+ (1) (2)))))", Fraction(3, 256)),
                ("(+ (1) (+ (1) (+ (1) (+ (1) (+ (1) (2))))))", Fraction(3, 1024)),
            },
        )

    def test_enumeration_from_dsl_with_variables_uniform(self):
        result = self.enumerate_dsl(
            fam_with_vars, fam_with_vars.uniform(), min_likelihood=-6
        )

        self.assertEqual(
            result,
            {
                ("(1)", Fraction(1, 4)),
                ("(2)", Fraction(1, 4)),
                ("(+ (1) (1))", Fraction(1, 64)),
                ("(+ (2) (1))", Fraction(1, 64)),
                ("(+ (1) (2))", Fraction(1, 64)),
                ("(+ (2) (2))", Fraction(1, 64)),
                ("(call (lam ($0_0)) (1))", Fraction(1, 80)),
                ("(call (lam (1)) (1))", Fraction(1, 80)),
                ("(call (lam (2)) (1))", Fraction(1, 80)),
                ("(call (lam ($0_0)) (2))", Fraction(1, 80)),
                ("(call (lam (1)) (2))", Fraction(1, 80)),
                ("(call (lam (2)) (2))", Fraction(1, 80)),
            },
        )

    def test_enumeration_from_dsl_with_ordering(self):
        result = self.enumerate_dsl(
            fam_with_ordering, fam_with_ordering.uniform(), min_likelihood=-6
        )
        self.assertEqual(result, {("(+ (1) (2) (3))", Fraction(1))})

    def test_enumeration_from_dsl_with_ordering_231(self):
        result = self.enumerate_dsl(
            fam_with_ordering_231, fam_with_ordering_231.uniform(), min_likelihood=-6
        )
        self.assertEqual(result, {("(+ (2) (3) (1))", Fraction(1))})



import unittest

import neurosym as ns


class EnumerationRegressionTest(unittest.TestCase):
    def assertRenderingEqual(self, actual, expected):
        print(actual)
        self.assertEqual(
            {line.strip() for line in actual.strip().split("\n")},
            {line.strip() for line in expected.strip().split("\n")},
        )

    def rendered_dsl(self, lambdas_kwargs=None):
        if lambdas_kwargs is None:
            lambdas_kwargs = {}
        dslf = ns.DSLFactory()
        dslf.known_types("i")
        dslf.lambdas(**lambdas_kwargs)
        return dslf.finalize().render()

    def test_basic(self):
        self.assertRenderingEqual(
            self.rendered_dsl(),
            """
            lam_0 :: L<#body|(i, i) -> i;(i, i) -> i> -> ((i, i) -> i, (i, i) -> i) -> #body
            lam_1 :: L<#body|(i, i) -> i;i -> i> -> ((i, i) -> i, i -> i) -> #body
            lam_2 :: L<#body|(i, i) -> i;i> -> ((i, i) -> i, i) -> #body
            lam_3 :: L<#body|(i, i) -> i> -> ((i, i) -> i) -> #body
            lam_4 :: L<#body|i -> i;(i, i) -> i> -> (i -> i, (i, i) -> i) -> #body
            lam_5 :: L<#body|i -> i;i -> i> -> (i -> i, i -> i) -> #body
            lam_6 :: L<#body|i -> i;i> -> (i -> i, i) -> #body
            lam_7 :: L<#body|i -> i> -> (i -> i) -> #body
            lam_8 :: L<#body|i;(i, i) -> i> -> (i, (i, i) -> i) -> #body
            lam_9 :: L<#body|i;i -> i> -> (i, i -> i) -> #body
            lam_10 :: L<#body|i;i> -> (i, i) -> #body
            lam_11 :: L<#body|i> -> i -> #body
            $0_0 :: V<(i, i) -> i@0>
            $1_0 :: V<(i, i) -> i@1>
            $2_0 :: V<(i, i) -> i@2>
            $3_0 :: V<(i, i) -> i@3>
            $0_1 :: V<i -> i@0>
            $1_1 :: V<i -> i@1>
            $2_1 :: V<i -> i@2>
            $3_1 :: V<i -> i@3>
            $0_2 :: V<i@0>
            $1_2 :: V<i@1>
            $2_2 :: V<i@2>
            $3_2 :: V<i@3>
            """,
        )

    def test_limited_arity(self):
        self.assertRenderingEqual(
            self.rendered_dsl(lambdas_kwargs=dict(max_arity=1)),
            """
            lam_0 :: L<#body|i -> i> -> (i -> i) -> #body
            lam_1 :: L<#body|i> -> i -> #body
            $0_0 :: V<i -> i@0>
            $1_0 :: V<i -> i@1>
            $2_0 :: V<i -> i@2>
            $3_0 :: V<i -> i@3>
            $0_1 :: V<i@0>
            $1_1 :: V<i@1>
            $2_1 :: V<i@2>
            $3_1 :: V<i@3>
            """,
        )

    def test_limited_type_depth(self):
        self.assertRenderingEqual(
            self.rendered_dsl(lambdas_kwargs=dict(max_type_depth=3.5)),
            """
            lam_0 :: L<#body|i -> i> -> (i -> i) -> #body
            lam_1 :: L<#body|i;i> -> (i, i) -> #body
            lam_2 :: L<#body|i> -> i -> #body
            $0_0 :: V<i -> i@0>
            $1_0 :: V<i -> i@1>
            $2_0 :: V<i -> i@2>
            $3_0 :: V<i -> i@3>
            $0_1 :: V<i@0>
            $1_1 :: V<i@1>
            $2_1 :: V<i@2>
            $3_1 :: V<i@3>
            """,
        )

    def test_limited_type_depth_env_depth(self):
        self.assertRenderingEqual(
            self.rendered_dsl(lambdas_kwargs=dict(max_type_depth=3.5, max_env_depth=2)),
            """
            lam_0 :: L<#body|i -> i> -> (i -> i) -> #body
            lam_1 :: L<#body|i;i> -> (i, i) -> #body
            lam_2 :: L<#body|i> -> i -> #body
            $0_0 :: V<i -> i@0>
            $1_0 :: V<i -> i@1>
            $0_1 :: V<i@0>
            $1_1 :: V<i@1>
            """,
        )



from abc import ABC, abstractmethod
from typing import Callable, List

import numpy as np

from neurosym.program_dist.tree_distribution.preorder_mask.undos import chain_undos


class PreorderMask(ABC):
    """
    Represents a mask on symbols that is updated during a preorder traversal of the tree.

    This can be used to implement such features as type checking and other constraints
        during enumeration, sampling, likelihood computation, and counting.
    """

    def __init__(self, tree_dist):
        self.tree_dist = tree_dist

    @abstractmethod
    def compute_mask(self, position: int, symbols: List[int]) -> List[bool]:
        """
        Compute the mask at the current position with the current symbols.

        The mask should be True for symbols that are allowed and False for symbols that are not.
        """

    @abstractmethod
    def on_entry(self, position: int, symbol: int) -> Callable[[], None]:
        """
        Called when entering a node in the preorder traversal.

        This can be used to update the mask.

        Returns a function that can be called to undo the changes made by this function.
        """

    @abstractmethod
    def on_exit(self, position: int, symbol: int) -> Callable[[], None]:
        """
        Called when exiting a node in the preorder traversal.

        This can be used to update the mask.

        Returns a function that can be called to undo the changes made by this function.
        """


class NoopPreorderMask(PreorderMask):
    """
    A mask that does nothing.
    """

    def compute_mask(self, position: int, symbols: List[int]) -> List[bool]:
        return [True] * len(symbols)

    def on_entry(self, position: int, symbol: int) -> Callable[[], None]:
        return lambda: None

    def on_exit(self, position: int, symbol: int) -> Callable[[], None]:
        return lambda: None


class ConjunctionPreorderMask(PreorderMask):
    """
    A mask that is the conjunction of multiple masks.
    """

    def __init__(self, tree_dist, masks):
        super().__init__(tree_dist)
        self.masks = masks

    def compute_mask(self, position: int, symbols: List[int]) -> List[bool]:
        symbols = np.array(symbols)
        mask = np.ones(len(symbols), dtype=bool)
        for m in self.masks:
            [valid_symbol_idxs] = np.where(mask)
            mask[valid_symbol_idxs] &= m.compute_mask(
                position, symbols[valid_symbol_idxs]
            )
        return mask.tolist()

    def on_entry(self, position: int, symbol: int) -> Callable[[], None]:
        undos = []
        for mask in self.masks:
            undos.append(mask.on_entry(position, symbol))
        return chain_undos(undos)

    def on_exit(self, position: int, symbol: int) -> Callable[[], None]:
        undos = []
        for mask in self.masks:
            undos.append(mask.on_exit(position, symbol))
        return chain_undos(undos)



from neurosym.program_dist.tree_distribution.ordering import DictionaryNodeOrdering
from neurosym.python_dsl.python_ast_tools import fields_for_node


def field_order(node, fields):
    """
    Get the order of the given fields in a node.
    """
    node_fields = fields_for_node(node)
    assert set(fields) == set(node_fields)
    return [node_fields.index(f) for f in fields]


def python_ordering_dictionary():
    """
    The ordering dictionary for Python nodes. This is a dictionary that maps node types
        to a list of integers. The integers are the indices of the fields in the node
        that should be visited in order.
    """
    fields = [
        ("ListComp~E", ["generators", "elt"]),
        ("GeneratorExp~E", ["generators", "elt"]),
        ("SetComp~E", ["generators", "elt"]),
        ("DictComp~E", ["generators", "key", "value"]),
    ]

    result = {}
    for node, fields in fields:
        result[node] = field_order(node, fields)
    return result


class PythonNodeOrdering(DictionaryNodeOrdering):
    """
    Orders the subnodes of a node according to a dictionary.
    """

    def __init__(self, dist):
        super().__init__(dist, python_ordering_dictionary(), tolerate_missing=True)



import unittest

import numpy as np

import neurosym as ns


class DSLSmoothingTest(unittest.TestCase):
    def setUp(self):
        self.dfa = None
        self.fam = None
        self.dist = None

    def dist_for_smoothing(
        self, programs, extra_programs=(), root="M", do_smooth_masking=False
    ):
        programs = [ns.python_to_python_ast(program) for program in programs]
        extra_programs = [
            ns.python_to_python_ast(program) for program in extra_programs
        ]
        dfa = ns.python_dfa()
        dsl = ns.create_python_dsl(
            dfa,
            ns.PythonDSLSubset.from_programs(
                dfa, *programs, *extra_programs, root=root
            ),
            root,
        )
        if do_smooth_masking:
            dsl_subset = ns.create_python_dsl(
                dfa,
                ns.PythonDSLSubset.from_programs(dfa, *programs, root=root),
                root,
            )
            smooth_mask = dsl.create_smoothing_mask(dsl_subset)
        else:
            smooth_mask = None
        fam = ns.BigramProgramDistributionFamily(
            dsl,
            additional_preorder_masks=[
                lambda dist, dsl: ns.python_def_use_mask.DefUseChainPreorderMask(
                    dist, dsl, ns.python_def_use_mask.DefUseMaskConfiguration(dfa, {})
                )
            ],
            node_ordering=ns.python_def_use_mask.PythonNodeOrdering,
        )
        counts = fam.count_programs(
            [
                [
                    ns.to_type_annotated_ns_s_exp(program, dfa, root)
                    for program in programs
                ]
            ]
        )
        dist = fam.counts_to_distribution(counts)[0]
        dist = dist.bound_minimum_likelihood(1e-6, smooth_mask)
        self.dfa = dfa
        self.fam = fam
        self.dist = dist

    def likelihood(self, program):
        program = ns.python_to_python_ast(program)
        log_p = self.fam.compute_likelihood(
            self.dist,
            ns.to_type_annotated_ns_s_exp(program, self.dfa, "M"),
        )
        return log_p

    def assertLikeli(self, prog, reciprocal_prob):
        self.assertAlmostEqual(self.likelihood(prog), -np.log(reciprocal_prob), 3)

    def standard_checks(self):
        # 1/2 [2], 1/2 [5]
        self.assertLikeli("x = 2 + 5", 2 * 2)
        # 1/2 [2], 1/1e6 [+], 1/2 [5], 1/2 [5]
        self.assertLikeli("x = 2 + (2 + 5)", 2 * 10**6 * 2 * 2)

    def test_smoothing_no_extra(self):
        self.dist_for_smoothing(["x = 2 + 5"])
        self.standard_checks()

    def test_smoothing_with_extra(self):
        self.dist_for_smoothing(["x = 2 + 5"], ["x = 2 * 3"])
        self.standard_checks()
        # 1/1e6 [*], 1/2 [2], 1/2 [5]
        self.assertLikeli("x = 2 * 5", 10**6 * 2 * 2)

    def test_smoothing_with_extra_subset_mask(self):
        self.dist_for_smoothing(["x = 2 + 5"], ["x = 2 * 3"], do_smooth_masking=True)
        self.standard_checks()
        # 1/1e6 [*], 1/2 [2], 1/2 [5]
        self.assertLikeli("x = 2 * 5", np.inf)



import itertools
from typing import Union

import stitch_core

from ..dsl.abstraction import AbstractionIndexParameter, AbstractionProduction
from ..programs.s_expression import SExpression
from ..programs.s_expression_render import (
    parse_s_expression,
    render_s_expression,
    symbols_for_program,
)
from ..types.type_signature import FunctionTypeSignature


def compute_abstraction_production(
    dsl,
    s_expression_using: SExpression,
    abstr_name: str,
    abstr_body: SExpression,
):
    """
    Compute the type of an abstraction production.

    Args:
        dsl: The DSL.
        s_expression_using: An SExpression using the abstraction.
        abstr: The original abstraction.

    Returns an AbstractionProduction corresponding to the abstraction.
    """
    abstr_body = inject_parameters(abstr_body)
    usage = next(x for x in s_expression_using.postorder if x.symbol == abstr_name)
    type_arguments = [dsl.compute_type(x) for x in usage.children]
    type_out = dsl.compute_type(
        abstr_body,
        lambda x: type_arguments[x.index]
        if isinstance(x, AbstractionIndexParameter)
        else None,
    ).typ
    type_signature = FunctionTypeSignature([x.typ for x in type_arguments], type_out)

    return AbstractionProduction(abstr_name, type_signature, abstr_body)


def inject_parameters(s_expression: Union[SExpression, str]):
    """
    Inject parameters into an SExpression, replacing leaves of the form #N with
    AbstractionIndexParameter(N).
    """
    if isinstance(s_expression, str):
        assert s_expression.startswith("#")
        return AbstractionIndexParameter(int(s_expression[1:]))
    return SExpression(
        s_expression.symbol,
        tuple(inject_parameters(x) for x in s_expression.children),
    )


def next_symbol(dsl):
    """
    Next free symbol of the form __N0 in the DSL.

    Returns __N, the prefix.
    """
    possible_conflict = [
        x[2:-1] for x in dsl.symbols() if x.startswith("__") and x.endswith("0")
    ]
    possible_conflict = {int(x) for x in possible_conflict if x.isnumeric()}
    number = next(x for x in itertools.count(1) if x not in possible_conflict)
    return f"__{number}"


def multi_lambda_to_single_lambda(dsl):
    lams = [x for x in dsl.productions if x.base_symbol() == "lam"]
    if not lams:
        return 0, {}
    max_index = max(x.get_index() for x in lams) + 1
    multi_to_single = {}
    zero_arg_lambda = None
    for lam in lams:
        if lam.arity == 0:
            assert zero_arg_lambda is None
            zero_arg_lambda = lam.get_index()
        if lam.arity == 1:
            continue
        multi_to_single[lam.get_index()] = [max_index + i for i in range(lam.arity)]
        max_index += lam.arity
    return zero_arg_lambda, multi_to_single


class StitchLambdaRewriter:
    """
    Rewrites a DSL to use only single-argument lambdas.

    Zero-argument lambdas are sent to a special symbol.

    Multi-argument lambdas are rewritten to multi single-argument lambdas.
    """

    def __init__(self, dsl):
        self.dsl = dsl
        (
            self.zero_arg_lambda_index_original,
            self.multi_to_single,
        ) = multi_lambda_to_single_lambda(dsl)
        self.zero_arg_lambda_symbol = next_symbol(dsl)

        self.first_single_to_multi = {
            single[0]: multi for multi, single in self.multi_to_single.items()
        }

    def to_stitch(self, s_exp):
        if isinstance(s_exp, str):
            return s_exp
        children = tuple(self.to_stitch(x) for x in s_exp.children)
        if not s_exp.symbol.startswith("lam"):
            return SExpression(
                s_exp.symbol,
                children,
            )
        prod = self.dsl.get_production(s_exp.symbol)
        if prod.arity == 0:
            return SExpression(
                self.zero_arg_lambda_symbol,
                children,
            )
        if prod.arity == 1:
            return SExpression(
                s_exp.symbol,
                children,
            )
        assert prod.arity > 1
        [result] = children
        for i in reversed(range(prod.arity)):
            result = SExpression(
                f"lam_{self.multi_to_single[prod.get_index()][i]}",
                (result,),
            )
        return result

    def from_stitch(self, s_exp):
        if isinstance(s_exp, str):
            return s_exp
        children = lambda: tuple(self.from_stitch(x) for x in s_exp.children)
        if s_exp.symbol == self.zero_arg_lambda_symbol:
            return SExpression(
                f"lam_{self.zero_arg_lambda_index_original}",
                children(),
            )
        if s_exp.symbol in self.dsl.symbols() or not s_exp.symbol.startswith("lam"):
            return SExpression(s_exp.symbol, children())
        index = int(s_exp.symbol[4:])
        original = self.first_single_to_multi[index]
        for i in range(len(self.multi_to_single[original])):
            assert s_exp.symbol == f"lam_{self.multi_to_single[original][i]}"
            [s_exp] = s_exp.children
        return SExpression(
            f"lam_{original}",
            (self.from_stitch(s_exp),),
        )


def single_step_compression(dsl, programs):
    """
    Run a single step of compression on a list of programs.
    """
    rewriter = StitchLambdaRewriter(dsl)
    programs = [rewriter.to_stitch(prog) for prog in programs]
    rendered = [render_s_expression(prog, for_stitch=True) for prog in programs]
    res = stitch_core.compress(
        rendered,
        1,
        no_curried_bodies=True,
        no_curried_metavars=True,
        abstraction_prefix=next_symbol(dsl),
    )
    abstr = res.abstractions[-1]
    rewritten = [
        rewriter.from_stitch(
            parse_s_expression(x, should_not_be_leaf={abstr.name}, for_stitch=True)
        )
        for x in res.rewritten
    ]
    user = next(x for x in rewritten if abstr.name in symbols_for_program(x))
    abstr_body = rewriter.from_stitch(
        parse_s_expression(abstr.body, should_not_be_leaf={abstr.name}, for_stitch=True)
    )
    prod = compute_abstraction_production(dsl, user, abstr.name, abstr_body)
    dsl2 = dsl.add_production(prod)
    return dsl2, rewritten


def multi_step_compression(dsl, programs, iterations):
    """
    Run multiple steps of compression on a list of programs.
    """
    for _ in range(iterations):
        dsl, programs = single_step_compression(dsl, programs)
    return dsl, programs



import ast
from typing import Dict, List

from frozendict import frozendict

from .python_ast_tools import fields_for_node

# exclude these tags from the dfa. these are all python 3.10+ features,
# and for consistency across python versions, we exclude them. We can
# add them back in later if we want to support them.
excluded_python_tags = [
    # match
    "Match",
    "MatchAs",
    "MatchMapping",
    "MatchSequence",
    "MatchValue",
    "MatchClass",
    "MatchOr",
    "MatchSingleton",
    "MatchStar",
    "match_case",
    "pattern",
    # trystar
    "TryStar",
]


default_transition_dict = frozendict(
    {
        "M": {ast.Module: {"body": "seqS", "type_ignores": "[TI]"}},
        "S": {
            (ast.FunctionDef, ast.AsyncFunctionDef, ast.ClassDef): {
                "body": "seqS",
                "decorator_list": "[E]",
                "bases": "[E]",
                "name": "Name",
                "args": "As",
                "returns": "TA",
                "type_comment": "TC",
                "keywords": "[K]",
            },
            ast.Return: {"value": "E"},
            ast.Delete: {"targets": "[L]"},
            (ast.Assign, ast.AugAssign, ast.AnnAssign): {
                "value": "E",
                "targets": "[L]",
                "target": "L",
                "type_comment": "TC",
                "op": "O",
                "annotation": "TA",
                "simple": "bool",
            },
            (
                ast.For,
                ast.AsyncFor,
                ast.While,
                ast.If,
                ast.With,
                ast.AsyncWith,
                ast.Try,
            ): {
                "iter": "E",
                "test": "E",
                "body": "seqS",
                "orelse": "seqS",
                "finalbody": "seqS",
                "items": "[W]",
                "handlers": "[EH]",
                "target": "L",
                "type_comment": "TC",
            },
            #         ast.Match: {"subject": "E", "cases": "C"},
            ast.Raise: {all: "E"},
            ast.Assert: {all: "E"},
            (ast.Import, ast.ImportFrom): {
                "names": "[alias]",
                "module": "NullableNameStr",
                "level": "int",
            },
            (ast.Global, ast.Nonlocal): {"names": "[NameStr]"},
            ast.Expr: {"value": "E"},
        },
        "E": {
            (ast.BoolOp, ast.BinOp, ast.UnaryOp, ast.Compare): {
                "op": "O",
                "ops": "[O]",
                "comparators": "[E]",
                "values": "[E]",
                all: "E",
            },
            ast.NamedExpr: {"value": "E", "target": "L"},
            ast.Lambda: {"args": "As", "body": "E"},
            (
                ast.IfExp,
                ast.Dict,
                ast.Set,
                ast.List,
                ast.Tuple,
                ast.Await,
                ast.Yield,
                ast.YieldFrom,
            ): {
                "ctx": "Ctx",
                "elts": "[StarredRoot]",
                "keys": "[E]",
                "values": "[E]",
                all: "E",
            },
            (ast.ListComp, ast.SetComp, ast.DictComp, ast.GeneratorExp): {
                "generators": "[C]",
                all: "E",
            },
            ast.Call: {
                "keywords": "[K]",
                "args": "[StarredRoot]",
                all: "E",
            },
            ast.JoinedStr: {"values": "[F]"},
            ast.Constant: {"value": "Const", "kind": "ConstKind"},
            ast.Name: {"id": "Name", "ctx": "Ctx"},
            (ast.Attribute, ast.Subscript, ast.Starred): {
                "value": "E",
                "attr": "NameStr",
                "slice": "SliceRoot",
                "ctx": "Ctx",
            },
        },
        "SliceRoot": {
            "_slice_content": {all: "E"},
            "_slice_slice": {all: "Slice"},
            "_slice_tuple": {all: "SliceTuple"},
        },
        "SliceTuple": {
            ast.Tuple: {"elts": "[SliceRoot]", "ctx": "Ctx"},
        },
        "[SliceRoot]": {"list": "SliceRoot"},
        "StarredRoot": {
            "_starred_content": {all: "E"},
            "_starred_starred": {all: "Starred"},
        },
        "Starred": {ast.Starred: {"value": "E", "ctx": "Ctx"}},
        "Slice": {
            ast.Slice: {"lower": "E", "upper": "E", "step": "E"},
        },
        "As": {
            ast.arguments: {
                ("kw_defaults", "defaults"): "[E]",
                ("args", "posonlyargs", "kwonlyargs"): "[A]",
                ("vararg", "kwarg"): "A",
            }
        },
        "A": {
            ast.arg: {"annotation": "TA", "arg": "Name", "type_comment": "TC"},
        },
        "F": {
            ast.FormattedValue: {"value": "E", "format_spec": "F", "conversion": "int"},
            ast.Constant: {all: "F"},
            ast.JoinedStr: {"values": "[F]"},
        },
        "C": {
            ast.comprehension: {
                "target": "L",
                "iter": "E",
                "ifs": "[E]",
                "is_async": "bool",
            }
        },
        "K": {ast.keyword: {"value": "E", "arg": "NullableNameStr"}},
        "EH": {
            ast.ExceptHandler: {
                "type": "E",
                "name": "NullableName",
                "body": "seqS",
            }
        },
        "W": {
            ast.withitem: {
                "context_expr": "E",
                "optional_vars": "L",
            }
        },
        "L": {
            ast.Name: {
                "id": "Name",
                "ctx": "Ctx",
            },
            ast.Tuple: {"elts": "[L]", "ctx": "Ctx"},
            ast.List: {"elts": "[L]", "ctx": "Ctx"},
            ast.Subscript: {"value": "E", "slice": "SliceRoot", "ctx": "Ctx"},
            ast.Attribute: {"value": "E", "attr": "NameStr", "ctx": "Ctx"},
            ast.Starred: {"value": "L", "ctx": "Ctx"},
            "_starred_content": {all: "L"},
            "_starred_starred": {all: "L"},
        },
        "seqS": {},
        "[E]": {"list": "E"},
        "[StarredRoot]": {"list": "StarredRoot"},
        "alias": {
            ast.alias: {
                "name": "NameStr",
                "asname": "NullableNameStr",
            },
        },
        "[NameStr]": {"list": "NameStr"},
        "TA": {all: {all: "TA"}, "list": "TA"},
        "[F]": {"list": "F"},
        "[A]": {"list": "A"},
        "[C]": {"list": "C"},
        "[EH]": {"list": "EH"},
        "[K]": {"list": "K"},
        "[L]": {"list": "L"},
        "[O]": {"list": "O"},
        "[W]": {"list": "W"},
        "[alias]": {"list": "alias"},
        "[TI]": {"list": "TI"},
    }
)


def python_dfa(
    transitions=default_transition_dict,
) -> Dict[str, Dict[str, List[str]]]:
    """
    Export a Discrete Tree Finite Automaton for the Python AST,
        in the form of a dictionary of the form dict[state, dict[tag, list[state]]].
    """
    all_tags = [
        x
        for x in dir(ast)
        if isinstance(getattr(ast, x), type)
        and issubclass(getattr(ast, x), ast.AST)
        and x not in excluded_python_tags
    ]

    extras = [
        "_slice_content",
        "_slice_slice",
        "_slice_tuple",
        "_starred_content",
        "_starred_starred",
    ]

    result = {}
    for state in transitions:
        result[state] = {}
        for tag in all_tags:
            t = getattr(ast, tag)
            out = compute_transition(transitions, state, t, fields_for_node(t))
            if out is not None:
                result[state][tag] = out
        for tag in extras:
            out = compute_transition(transitions, state, tag, [None])
            if out is not None:
                result[state][tag] = out

        missing = (
            set(all_types_as_string(list(transitions[state])))
            - set(result[state])
            - {"list", "/seq", "/splice"}
        )
        if missing:
            raise RuntimeError(f"in state {state}: missing {missing}")

        if "list" in transitions[state]:
            result[state]["list"] = [transitions[state]["list"]]
    result["seqS"]["/seq"] = ["S"]
    result["seqS"]["/subseq"] = ["S"]
    result["seqS"]["/choiceseq"] = ["S"]
    result["S"]["/splice"] = ["seqS"]
    return result


def compute_transition(transitions, state, typ, fields):
    """
    Compute the list of states that the DFA should transition to
        for each child of a node of type `typ` in state `state`.

    :param transitions: the transition dictionary
    :param state: the current state
    :param typ: the type of the node
    :param fields: the fields of the node

    :return: a list of states to transition to, or None if
        the node could not be matched.
    """
    transition = transitions[state]
    transition = compute_match(transition, typ, default=False)
    if transition is False:
        return None
    return [compute_match(transition, field) for field in fields]


def all_types_as_string(ts):
    """
    Converts all the types in the given set into strings.
        E.g., (ast.Name, ast.If) -> ["Name", "If"]
    """
    if isinstance(ts, (list, tuple)):
        for x in ts:
            yield from all_types_as_string(x)
        return
    if ts is all:
        return
    if isinstance(ts, type):
        yield ts.__name__
        return
    assert isinstance(ts, str), ts
    yield ts


def compute_match(transition, key, default=None):
    """
    Compute the match for the given key in the transition dictionary. Handles tuples
        and the special case where the key is `all`.

    :param transition: the transition dictionary
    :param key: the key to look up
    :param default: the default value to return if the key is not found

    :return: the value corresponding to the key in the transition dictionary
    """
    for k, v in transition.items():
        if k is all:
            return v
        if not isinstance(k, tuple):
            k = (k,)
        if key in k:
            return v
    if default is not None:
        return default
    raise RuntimeError(f"could not find {key} in {transition}")



def assertDSL(test_obj, dsl, expected):
    dsl = "\n".join(sorted([line.strip() for line in dsl.split("\n") if line.strip()]))
    expected = "\n".join(
        sorted([line.strip() for line in expected.split("\n") if line.strip()])
    )
    print(dsl)
    test_obj.maxDiff = None
    test_obj.assertEqual(dsl, expected)



# Let c be the chunk size, h be the likelihood of the program we are trying to
# find, and T be the number of iterations we need to do.
#
# We have that the total time taken is
#     sum_{t=0}^{T - 1} e^{c * t} + f * e^{c * T}
# where f is the fraction of the last chunk. We can compute this in expectation
# as 1/2, which leads us to
#     sum_{t=0}^{T - 1} e^{c * t} + e^{c * T} / 2
# which can be computed as
#     (e^{c * (T + 1)} - 1) / (e^c - 1) - e^{c * T} / 2
# which can be approximated closely as
#     e^{cT} * e^c / (e^c - 1) - e^{c * T} / 2
# which is equal to
#     e^{cT} * (e^c / (e^c - 1) - 1/2)
# we then know that T = ceil(h / c), and letting g = T - h / c, we have
#     e^{cT} = e^{c * (h / c + g)} = e^h * e^{c * g}
# if we assume g is uniformly distributed between 0 and 1, we have
#     e^{c * g} = (e^c - 1) / c
# we then have that the expected time is
#     (e^c - 1) / c * (e^c / (e^c - 1) - 1/2)
# which is optimized per wolfram at
#     W(1/e) + 1
# which is approximately 1.28.
DEFAULT_CHUNK_SIZE = 1.278



import neurosym as ns
from neurosym.examples import near


def assertDSLEnumerable(dsl, out_t, max_depth=5):
    t = ns.TypeDefiner(L=10, O=4)
    t.typedef("fL", "{f, $L}")
    t.typedef("fO", "{f, $O}")

    def checker(x):
        """Initialize and return True always"""
        x = x.program
        xx = dsl.compute(dsl.initialize(x))
        print(xx)
        return True

    g = near.near_graph(dsl, t(out_t), is_goal=checker)

    def cost(x):
        if isinstance(x.program, ns.SExpression) and x.program.children:
            return len(str(x.program.children[0]))
        return 0

    # should not raise StopIteration.
    for _ in ns.search.bounded_astar(g, cost, max_depth=max_depth):
        pass



import json
import unittest

import neurosym as ns


class TestDFARegression(unittest.TestCase):
    def test_python_dfa(self):
        with open("test_data/dfa.json") as f:
            dfa = json.load(f)
        self.assertEqual(ns.python_dfa(), dfa)



import copy
import unittest
from textwrap import dedent

from parameterized import parameterized

import neurosym as ns

from .utils import fit_to, small_set_runnable_code_examples


class DefUseMaskTestGeneric(unittest.TestCase):
    def get_handler_except_mask(self, handler):
        return {
            k: v for k, v in handler.__dict__.items() if k not in {"mask", "config"}
        }

    def get_handlers_except_mask(self, mask):
        handlers = mask.masks[-1].handlers
        return [self.get_handler_except_mask(handler) for handler in handlers]

    def annotate_program(self, program, with_undo, with_undo_exit):
        def replace_node_midstream(s_exp, mask, position, alts):
            if with_undo:
                for alt in alts:
                    undo = mask.on_entry(position, alt)
                    undo()
            if with_undo_exit:
                for alt in alts:
                    undo_entry = mask.on_entry(position, alt)
                    print("*" * 80)
                    print("handlers", mask.masks[-1].handlers)
                    last_handler = copy.deepcopy(self.get_handlers_except_mask(mask))
                    print("copy", last_handler)
                    undo_exit = mask.on_exit(position, alt)
                    print("after exit", self.get_handlers_except_mask(mask))
                    undo_exit()
                    print(
                        "after undo exit",
                        self.get_handlers_except_mask(mask),
                    )
                    self.assertEqual(
                        last_handler,
                        self.get_handlers_except_mask(mask),
                    )
                    undo_entry()

            return s_exp

        dfa, _, fam, _ = fit_to(
            [program], parser=ns.python_to_python_ast, include_type_preorder_mask=False
        )
        td = fam.tree_distribution_skeleton
        result = list(
            ns.collect_preorder_symbols(
                ns.to_type_annotated_ns_s_exp(
                    ns.python_to_python_ast(program), dfa, "M"
                ),
                fam.tree_distribution_skeleton,
                replace_node_midstream=replace_node_midstream,
            )
        )
        result = [
            (ns.render_s_expression(s_exp), [td.symbols[i][0] for i in alts])
            for s_exp, alts, _ in result
        ]
        # print(result)
        return result

    def assertUndoHasNoEffect(self, program):
        self.maxDiff = None
        with_undo = self.annotate_program(program, with_undo=True, with_undo_exit=False)
        with_undo_exit = self.annotate_program(
            program, with_undo=False, with_undo_exit=True
        )
        without_undo = self.annotate_program(
            program, with_undo=False, with_undo_exit=False
        )
        self.assertEqual(with_undo, without_undo)
        self.assertEqual(with_undo_exit, without_undo)


class DefUseMaskTest(DefUseMaskTestGeneric):
    def test_annotate_alternate_symbols(self):
        self.assertUndoHasNoEffect("x = 2; y = x; z = y")

    def test_for(self):
        self.assertUndoHasNoEffect(
            dedent(
                r"""
                for i in range(2):
                    x = 2
                print(i)
                """
            )
        )

    def test_for_then_subscript(self):
        self.assertUndoHasNoEffect(
            dedent(
                r"""
                for i in range(2):
                    x = 2
                x[0] = 2
                """
            )
        )

    def test_temp(self):
        self.assertUndoHasNoEffect(
            dedent(
                r"""
                for u in 2:
                    pass
                x[0] = 2
                """
            )
        )

    @parameterized.expand([(i,) for i in range(50)])
    def test_realistic(self, i):
        if i in {22, 31, 41, 57, 95, 100, 106, 109, 112, 114, 119, 181, 182}:
            # forward declaration of
            # input for 22/41/100/119
            # n for 31/57/112/114
            # m for 95
            # sp for 106
            # mp for 109
            # dp for 181
            # lo for 182 [this one's weird]
            return
        example = small_set_runnable_code_examples()[i]["solution"]
        print(example)
        self.assertUndoHasNoEffect(example)



import ast

from neurosym.python_dsl.names import PYTHON_DSL_SEPARATOR
from neurosym.types.type import AtomicType, ListType
from neurosym.types.type_string_repr import parse_type

# types we do not care about, so just send them to the same state
pruned_python_dfa_states = ["TA"]


def fields_for_node(node):
    """
    Get the fields for a node. If the node is a string, get the
        segment before the python dsl separator
    """
    if isinstance(node, str):
        node = node.split(PYTHON_DSL_SEPARATOR)[0]
        node = getattr(ast, node)

    return node._fields


def field_is_body(node_type: type, field_name: str):
    """
    Returns whether a field is a body field.

    Args:
        node_type: The type of the node.
        field_name: The name of the field.

    Returns:
        Whether the field is a body field.
    """
    assert isinstance(node_type, type)
    if node_type == ast.IfExp:
        return False  # not body of statements
    if node_type == ast.Lambda:
        return False  # not body of statements
    return field_name in {"body", "orelse", "finalbody"}


def field_is_starrable(node_type, field_name):
    """
    Field is starrable if it is a list of elements or a call with args.

    Args:
        node_type: The type of the node.
        field_name: The name of the field.

    Returns:
        Whether the field is starrable.
    """
    if field_name == "elts":
        assert node_type in {
            ast.List,
            ast.Tuple,
            ast.Set,
        }, (node_type, field_name)
        return True
    return node_type == ast.Call and field_name == "args"


def name_field(node: ast.AST):
    """
    Find the name field for a node.
    """
    t = type(node)
    if t == ast.Name:
        return "id"
    if t == ast.arg:
        return "arg"
    if t == ast.FunctionDef:
        return "name"
    if t == ast.AsyncFunctionDef:
        return "name"
    if t == ast.ExceptHandler:
        return "name" if node.name is not None else None
    if t == ast.ClassDef:
        return "name"
    if t == ast.alias:
        if node.asname is None:
            return "name"
        return "asname"
    raise NotImplementedError(f"Unexpected type: {t}")


def is_sequence_type(x):
    """
    Returns whether a type is a sequence type.
    Sequence types are either lists or seqS.
    """
    x = parse_type(x)
    if isinstance(x, ListType):
        return True
    if not isinstance(x, AtomicType):
        return False
    return x.name == "seqS"


def is_sequence_symbol(x):
    """
    Returns whether a symbol is a sequence symbol.
    """
    return x in ["/seq", "/subseq", "list", "/choiceseq"]


def is_sequence(type_name, head_symbol):
    """
    Returns whether a given type and head symbol correspond to a sequence.
        If there is a mismatch between the type and the head symbol, this function
        errors.

    Args:
        type_name: The type name.
        head_symbol: The head symbol.
    """

    seq_type = is_sequence_type(type_name)
    seq_symbol = is_sequence_symbol(head_symbol)
    if type_name in pruned_python_dfa_states:
        return seq_symbol
    return seq_type and seq_symbol


def clean_type(x):
    """
    Replace [] with __ in the type name
    """
    return x.replace("[", "_").replace("]", "_")


def unclean_type(x):
    """
    Replace __ with [] in the type name
    """
    if "_" not in x:
        return x
    assert x.count("_") == 2, x
    return x.replace("_", "[", 1).replace("_", "]", 1)



from abc import abstractmethod
from dataclasses import dataclass
from functools import cached_property
from types import NoneType
from typing import Callable, Dict, List, Tuple, Union

import numpy as np

from neurosym.dsl.dsl import ROOT_SYMBOL
from neurosym.program_dist.distribution import (
    ProgramDistribution,
    ProgramDistributionFamily,
)
from neurosym.program_dist.enumeration_chunk_size import DEFAULT_CHUNK_SIZE
from neurosym.program_dist.tree_distribution.ordering import NodeOrdering
from neurosym.program_dist.tree_distribution.preorder_mask.preorder_mask import (
    PreorderMask,
)
from neurosym.programs.s_expression import SExpression


@dataclass
class TreeDistribution:
    """
    Distribution over SExpressions as trees.

    Internally, we represent the productions in the language as integers, which we
        call indices.
    """

    limit: int
    # input: tuple of tuples of (ancestor index, position)
    #        which is the path to the current node, with the
    #        most immediate ancestor at the end.
    # output: list of (production index, likelihood) pairs
    distribution: Dict[Tuple[Tuple[int, int], ...], List[Tuple[int, float]]]
    # production index -> (symbol, arity). at 0 should be the root.
    symbols: List[Tuple[str, int]]
    # Preorder mask constructor
    mask_constructor: Callable[["TreeDistribution"], PreorderMask]
    # Node ordering
    node_ordering: Callable[["TreeDistribution"], NodeOrdering]

    @cached_property
    def symbol_to_index(self) -> Dict[str, int]:
        return {symbol: i for i, (symbol, _) in enumerate(self.symbols)}

    @cached_property
    def index_within_distribution_list(
        self,
    ) -> Dict[Tuple[Tuple[int, int], ...], Dict[int, int]]:
        return {
            k: {x: i for i, (x, _) in enumerate(v)}
            for k, v in self.distribution.items()
        }

    @cached_property
    def distribution_dict(self) -> Dict[Tuple[Tuple[int, int], ...], Dict[int, float]]:
        return {k: dict(v) for k, v in self.distribution.items()}

    @cached_property
    def likelihood_arrays(
        self,
    ) -> Dict[Tuple[Tuple[int, int], ...], Tuple[np.ndarray, np.ndarray]]:
        return {
            k: (
                np.array([x[0] for x in v]),
                np.array([x[1] for x in v]),
            )
            for k, v in self.distribution.items()
        }

    @cached_property
    def sampling_dict_arrays(
        self,
    ) -> Dict[Tuple[Tuple[int, int], ...], Tuple[np.ndarray, np.ndarray]]:
        return {
            k: (syms, np.exp(log_probs))
            for k, (syms, log_probs) in self.likelihood_arrays.items()
        }

    @cached_property
    def ordering(self) -> NodeOrdering:
        return self.node_ordering(self)


class TreeProgramDistributionFamily(ProgramDistributionFamily):
    """
    See `tree_dist_enumerator.py` for more information.
    """

    @abstractmethod
    def compute_tree_distribution(
        self, distribution: Union[ProgramDistribution, NoneType]
    ) -> TreeDistribution:
        """
        Returns a tree distribution representing the given program distribution.

        If `distribution` is `None`, returns a tree distribution with all fields
            initialized, but the probabilities are not guaranteed to have any properties.
            This is useful for tasks where you want the skeleton of the tree distribution,
            but don't need the actual distribution.
        """

    def tree_distribution(self, distribution: ProgramDistribution) -> TreeDistribution:
        """
        Cached version of `compute_tree_distribution`.
        """
        # This is a bit of a hack, but it reduces the need to pass around
        # the tree distribution everywhere, or to compute it multiple times.
        # pylint: disable=protected-access
        if not hasattr(distribution, "_tree_distribution"):
            distribution._tree_distribution = self.compute_tree_distribution(
                distribution
            )
        return distribution._tree_distribution

    @cached_property
    def tree_distribution_skeleton(self) -> TreeDistribution:
        """
        Cached version of `compute_tree_distribution(None)`.
        """

        return self.compute_tree_distribution(None)

    def enumerate(
        self,
        dist: ProgramDistribution,
        *,
        min_likelihood: float = float("-inf"),
        chunk_size: float = DEFAULT_CHUNK_SIZE,
    ):
        # pylint: disable=cyclic-import
        from neurosym.program_dist.tree_distribution.tree_dist_enumerator import (
            enumerate_tree_dist,
        )

        tree_dist = self.tree_distribution(dist)

        return enumerate_tree_dist(
            tree_dist, min_likelihood=min_likelihood, chunk_size=chunk_size
        )

    def compute_likelihood(
        self,
        dist: ProgramDistribution,
        program: SExpression,
        tracker: Union[NoneType, Callable[[SExpression, float], NoneType]] = None,
    ) -> float:
        """
        Compute the likelihood of a program under a distribution.
        """
        # pylint: disable=cyclic-import
        from .tree_dist_likelihood_computer import compute_likelihood

        dist = self.tree_distribution(dist)
        preorder_mask = dist.mask_constructor(dist)
        preorder_mask.on_entry(0, 0)
        return compute_likelihood(dist, program, ((0, 0),), preorder_mask, tracker)

    def compute_likelihood_per_node(
        self,
        dist: ProgramDistribution,
        program: SExpression,
    ) -> Dict[SExpression, float]:
        """
        Compute the likelihood of a program under a distribution.
        """
        likelihoods = []

        self.compute_likelihood(
            dist,
            program,
            lambda nodes, likelihood: likelihoods.append((nodes, likelihood)),
        )
        return likelihoods

    def sample(
        self,
        dist: ProgramDistribution,
        rng: np.random.RandomState,
        *,
        depth_limit=float("inf"),
    ) -> SExpression:
        # pylint: disable=cyclic-import

        from neurosym.program_dist.tree_distribution.tree_dist_sampler import (
            sample_tree_dist,
        )

        tree_dist = self.tree_distribution(dist)
        element = sample_tree_dist(tree_dist, rng, depth_limit=depth_limit)
        assert element.symbol == ROOT_SYMBOL
        [element] = element.children
        return element



import uuid
from abc import ABC, abstractmethod
from dataclasses import dataclass
from functools import cached_property
from typing import Callable, Dict, List, Tuple

import numpy as np


class Type(ABC):
    @abstractmethod
    def walk_type_nodes(self):
        """
        Walk the type tree and yield all nodes.
        """
        raise NotImplementedError

    def node_summary(self):
        """
        Return a summary of the node.
        """
        summary_dict = {
            "_type": self.__class__.__name__,
            **self.inherent_parameters(),
        }
        summary_dict = dict(sorted(summary_dict.items()))
        return str(summary_dict)

    @abstractmethod
    def inherent_parameters(self):
        """
        Return a list of inherent parameters of the type, i.e., all parameters not associated
            with the children.
        """
        raise NotImplementedError

    @abstractmethod
    def children(self):
        """
        Iterate over direct children and yield each.
        """
        raise NotImplementedError

    @abstractmethod
    def unify(
        self, other: "Type", already_tried_other_direction=False
    ) -> Dict[str, "Type"]:
        """
        Unify this type with another type. Returns a dictionary of substitutions.

        Raise UnificationError if the types cannot be unified.

        If already_tried_other_direction is True, then we have already tried to unify
        the other direction and failed, so we should not try again.
        """
        raise NotImplementedError

    @abstractmethod
    def subst_type_vars(self, subst: Dict[str, "Type"]):
        """
        Substitute type variables in the type according to the substitution.
        """
        raise NotImplementedError

    def is_atomic(self):
        """
        Return True if the type is atomic.
        """
        return len(list(self.walk_type_nodes())) == 1 and not isinstance(
            self, TypeVariable
        )

    def depth_of_type_variables(self) -> List[Tuple[str, float]]:
        """
        Return a list of tuples of the form (type_variable, depth) where depth is the
        depth of the type variable in the type tree.
        """
        node_height = np.log2(1 + len(list(self.children())))
        return [
            (tv, depth + node_height)
            for child in self.children()
            for tv, depth in child.depth_of_type_variables()
        ]

    def max_depth_per_type_variable(self) -> Dict[str, float]:
        """
        Return a dictionary mapping type variables to their maximum depth in the type
        tree.
        """
        result = {}
        for tv, depth in self.depth_of_type_variables():
            result[tv] = max(result.get(tv, 0), depth)
        return result

    def get_type_vars(self):
        """
        Return a set of all type variables in the type.
        """
        return sorted({tv for tv, _ in self.depth_of_type_variables()})

    def has_type_vars(self):
        """
        Return True if the type has type variables.
        """
        return len(self.get_type_vars()) > 0

    @cached_property
    def depth(self):
        """
        Return the depth of the type tree.
        """
        children = [child.depth for child in self.children()]
        depth = max(children + [0]) + np.log2(len(children) + 1)
        return depth


class UnificationError(Exception):
    pass


@dataclass(frozen=True, eq=True)
class AtomicType(Type):
    name: str

    def __post_init__(self):
        assert self.name.isidentifier(), f"{self.name} is not a valid identifier"

    def walk_type_nodes(self):
        yield self

    def inherent_parameters(self):
        return dict(name=self.name)

    def children(self):
        yield from []

    def unify(
        self, other: "Type", already_tried_other_direction=False
    ) -> Dict[str, "Type"]:
        if isinstance(other, AtomicType):
            if self.name != other.name:
                raise UnificationError(f"{self} != {other}")
            return {}
        if already_tried_other_direction:
            raise UnificationError(f"{self} != {other}")
        return other.unify(self, already_tried_other_direction=True)

    def subst_type_vars(self, subst: Dict[str, Type]):
        return self


@dataclass(frozen=True, eq=True)
class TensorType(Type):
    """
    A tensor is a type that represents a tensor of a given shape.
    """

    dtype: Type
    shape: Tuple[int]

    def walk_type_nodes(self):
        # is atomic. we do not consider the dtype and shape as nodes.
        yield self

    def inherent_parameters(self):
        return dict(dtype=self.dtype, shape=self.shape)

    def children(self):
        yield from []

    def subst_type_vars(self, subst: Dict[str, Type]):
        del subst
        assert not self.has_type_vars()
        return self

    def unify(
        self, other: "Type", already_tried_other_direction=False
    ) -> Dict[str, "Type"]:
        if isinstance(other, TensorType):
            if self.shape != other.shape:
                raise UnificationError(f"{self} != {other}")
            return self.dtype.unify(other.dtype)
        if already_tried_other_direction:
            raise UnificationError(f"{self} != {other}")
        return other.unify(self, already_tried_other_direction=True)


@dataclass(frozen=True, eq=True)
class ListType(Type):
    """
    A list type is a type of the form [t] where t is a type.
    """

    element_type: Type

    def walk_type_nodes(self):
        yield self
        yield from self.element_type.walk_type_nodes()

    def inherent_parameters(self):
        return dict()

    def children(self):
        yield self.element_type

    def subst_type_vars(self, subst: Dict[str, Type]):
        return ListType(self.element_type.subst_type_vars(subst))

    def unify(
        self, other: "Type", already_tried_other_direction=False
    ) -> Dict[str, "Type"]:
        if isinstance(other, ListType):
            return self.element_type.unify(other.element_type)
        if already_tried_other_direction:
            raise UnificationError(f"{self} != {other}")
        return other.unify(self, already_tried_other_direction=True)


@dataclass(frozen=True, eq=True)
class ArrowType(Type):
    """
    An arrow type is a type of the form t1 -> t2 where t1 and t2 are types.
    """

    input_type: Tuple[Type]
    output_type: Type

    def __post_init__(self):
        assert isinstance(self.input_type, tuple), "input_type must be a tuple"

    def subst_type_vars(self, subst: Dict[str, Type]):
        return ArrowType(
            tuple(t.subst_type_vars(subst) for t in self.input_type),
            self.output_type.subst_type_vars(subst),
        )

    def walk_type_nodes(self):
        yield self
        for t in self.input_type:
            yield from t.walk_type_nodes()
        yield from self.output_type.walk_type_nodes()

    def inherent_parameters(self):
        return dict()

    def children(self):
        yield from self.input_type
        yield self.output_type

    def unify(
        self, other: "Type", already_tried_other_direction=False
    ) -> Dict[str, "Type"]:
        if not isinstance(other, ArrowType):
            if already_tried_other_direction:
                raise UnificationError(f"{self} != {other}")
            return other.unify(self, already_tried_other_direction=True)
        if len(self.input_type) != len(other.input_type):
            raise UnificationError(f"{self} != {other}")
        individuals = []
        for t1, t2 in zip(self.input_type, other.input_type):
            individuals.append(t1.unify(t2))
        individuals.append(self.output_type.unify(other.output_type))
        subst = {}
        for individual in individuals:
            for k, v in individual.items():
                if k in subst:
                    if subst[k] != v:
                        raise UnificationError(f"{k} is unified to {subst[k]} and {v}")
                else:
                    subst[k] = v
        return subst


@dataclass(frozen=True, eq=True)
class GenericTypeVariable(Type):
    name: str

    @classmethod
    def fresh(cls):
        return cls(f"_{uuid.uuid4().hex}")

    def __post_init__(self):
        assert self.name.isidentifier(), f"{self.name} is not a valid identifier"

    def depth_of_type_variables(self) -> List[Tuple[str, int]]:
        return [(self.name, 0)]

    def subst_type_vars(self, subst: Dict[str, Type]):
        return subst.get(self.name, self)

    def walk_type_nodes(self):
        yield self

    def inherent_parameters(self):
        return dict()

    def children(self):
        yield from []

    def unify(
        self, other: "Type", already_tried_other_direction=False
    ) -> Dict[str, "Type"]:
        return {self.name: other}


class TypeVariable(GenericTypeVariable):
    pass


@dataclass(frozen=True, eq=True)
class FilteredTypeVariable(GenericTypeVariable):
    type_filter: Callable[[Type], bool]

    def unify(
        self, other: "Type", already_tried_other_direction=False
    ) -> Dict[str, "Type"]:
        if self.type_filter(other):
            return {self.name: other}
        raise UnificationError(f"{self} does not match {other}")


float_t = AtomicType("f")



import unittest
from fractions import Fraction

import numpy as np

import neurosym as ns

from .utils import fit_to


class EnumerateFittedDslTest(unittest.TestCase):
    def enumerate(self, *programs):
        _, _, fam, dist = fit_to(programs)
        out = [
            (
                Fraction.from_float(np.exp(y)).limit_denominator(),
                ns.s_exp_to_python(ns.render_s_expression(x)),
            )
            for x, y in fam.enumerate(dist, min_likelihood=-10)
        ]
        out = sorted(out, key=lambda x: (-x[0], x[1]))
        print(out)
        return out

    def test_enumerate_fitted_dsl_basic(self):
        self.assertEqual(
            self.enumerate("x = y + 2 + 2"),
            [
                (Fraction(1, 2), "x = y + 2"),
                (Fraction(1, 4), "x = y + 2 + 2"),
                (Fraction(1, 8), "x = y + 2 + 2 + 2"),
                (Fraction(1, 16), "x = y + 2 + 2 + 2 + 2"),
                (Fraction(1, 32), "x = y + 2 + 2 + 2 + 2 + 2"),
                (Fraction(1, 64), "x = y + 2 + 2 + 2 + 2 + 2 + 2"),
                (Fraction(1, 128), "x = y + 2 + 2 + 2 + 2 + 2 + 2 + 2"),
                (Fraction(1, 256), "x = y + 2 + 2 + 2 + 2 + 2 + 2 + 2 + 2"),
                (Fraction(1, 512), "x = y + 2 + 2 + 2 + 2 + 2 + 2 + 2 + 2 + 2"),
                (Fraction(1, 1024), "x = y + 2 + 2 + 2 + 2 + 2 + 2 + 2 + 2 + 2 + 2"),
                (
                    Fraction(1, 2048),
                    "x = y + 2 + 2 + 2 + 2 + 2 + 2 + 2 + 2 + 2 + 2 + 2",
                ),
                (
                    Fraction(1, 4096),
                    "x = y + 2 + 2 + 2 + 2 + 2 + 2 + 2 + 2 + 2 + 2 + 2 + 2",
                ),
                (
                    Fraction(1, 8192),
                    "x = y + 2 + 2 + 2 + 2 + 2 + 2 + 2 + 2 + 2 + 2 + 2 + 2 + 2",
                ),
                (
                    Fraction(1, 16384),
                    "x = y + 2 + 2 + 2 + 2 + 2 + 2 + 2 + 2 + 2 + 2 + 2 + 2 + 2 + 2",
                ),
            ],
        )

    def test_enumerate_fitted_dsl_cartesian(self):
        self.assertEqual(
            self.enumerate("x = 2", "y = 3"),
            [
                (Fraction(1, 4), "x = 2"),
                (Fraction(1, 4), "x = 3"),
                (Fraction(1, 4), "y = 2"),
                (Fraction(1, 4), "y = 3"),
            ],
        )

    def test_enumerate_def_use_check(self):
        self.assertEqual(
            self.enumerate("x = 2; y = x", "y = 2; x = y"),
            [
                (Fraction(1, 8), "x = 2\nx = 2"),
                (Fraction(1, 8), "x = 2\nx = x"),
                (Fraction(1, 8), "x = 2\ny = 2"),
                (Fraction(1, 8), "x = 2\ny = x"),
                (Fraction(1, 8), "y = 2\nx = 2"),
                (Fraction(1, 8), "y = 2\nx = y"),
                (Fraction(1, 8), "y = 2\ny = 2"),
                (Fraction(1, 8), "y = 2\ny = y"),
            ],
        )

    def test_enumerate_def_use_check_wglobal(self):
        self.assertEqual(
            self.enumerate("x = print; y = x", "y = print; x = y"),
            [
                (Fraction(1, 6), "x = print\nx = print"),
                (Fraction(1, 6), "x = print\ny = print"),
                (Fraction(1, 6), "y = print\nx = print"),
                (Fraction(1, 6), "y = print\ny = print"),
                (Fraction(1, 12), "x = print\nx = x"),
                (Fraction(1, 12), "x = print\ny = x"),
                (Fraction(1, 12), "y = print\nx = y"),
                (Fraction(1, 12), "y = print\ny = y"),
            ],
        )



import re

from neurosym.python_dsl.names import PYTHON_DSL_SEPARATOR

NAME_REGEX = re.compile(
    r"const-(?P<typ>&)(?P<name>\w+|\*):(?P<scope>\d+)"
    + PYTHON_DSL_SEPARATOR
    + r"(?P<dfa_sym>Name|NullableName|NameStr|NullableNameStr)$"
)
GLOBAL_REGEX = re.compile(
    r"const-(?P<typ>g)_(?P<name>\w+|\*)~(Name|NullableName|NameStr|NullableNameStr)"
)


def match_either_name_or_global(s):
    return NAME_REGEX.match(s) or GLOBAL_REGEX.match(s)



import ast
from typing import Callable, Dict, List, Union

from frozendict import frozendict

from neurosym.programs.s_expression import SExpression
from neurosym.programs.s_expression_render import render_s_expression
from neurosym.python_dsl.convert_python.parse_python import python_to_python_ast
from neurosym.python_dsl.convert_python.parse_s_exp import s_exp_to_python_ast
from neurosym.python_dsl.convert_python.python_ast import (
    NodeAST,
    PythonAST,
    SequenceAST,
)
from neurosym.python_dsl.run_dfa import add_disambiguating_type_tags


def python_to_s_exp(code: Union[str, ast.AST], **kwargs) -> str:
    """
    Converts python code to an s-expression.
    """
    return render_s_expression(python_to_python_ast(code).to_ns_s_exp(kwargs))


def s_exp_to_python(
    code: Union[str, SExpression],
    node_hooks: Dict[str, Callable[[str, List[PythonAST]], PythonAST]] = frozendict(),
) -> str:
    """
    Converts an s expression to python code.
    """
    return s_exp_to_python_ast(code, node_hooks).to_python()


def to_type_annotated_ns_s_exp(
    code: PythonAST, dfa: dict, start_state: str
) -> SExpression:
    """
    Like to_ns_s_exp, but adds type annotations.
    """
    return add_disambiguating_type_tags(
        dfa, code.to_ns_s_exp(dict(no_leaves=True)), start_state
    )


def python_statement_to_python_ast(code: Union[str, ast.AST]) -> PythonAST:
    """
    Like python_to_python_ast, but for a single statement.
    """
    code = python_statements_to_python_ast(code)
    assert (
        len(code.elements) == 1
    ), f"expected only one statement; got: [{[x.to_python() for x in code.elements]}]]"
    code = code.elements[0]
    return code


def python_statements_to_python_ast(code: Union[str, ast.AST]) -> SequenceAST:
    """
    Like python_to_python_ast, but for a sequence of statements.
    """
    code = python_to_python_ast(code)
    assert isinstance(code, NodeAST) and code.typ is ast.Module
    assert len(code.children) == 2
    code = code.children[0]
    assert isinstance(code, SequenceAST), code
    return code



from typing import Callable, Tuple

from neurosym.program_dist.tree_distribution.preorder_mask.undos import (
    chain_undos,
    remove_last_n_elements,
)

from ..handler import Handler
from .defining_statement_handler import ChildFrameCreatorHandler


class DefiningConstructHandler(ChildFrameCreatorHandler):
    """
    Handles defining constructs like function and class definitions.

    These are constructs that have a child frame defined as the body of the construct,
        as well as a name that is defined in the parent frame.
    """

    # these fields must be defined in the subclass
    construct_name_field: str = None

    def __init__(self, mask, defined_production_idxs, config):
        super().__init__(mask, defined_production_idxs, config)
        assert isinstance(self.construct_name_field, str)

    def on_child_enter(
        self, position: int, symbol: int
    ) -> Tuple[Handler, Callable[[], None]]:
        if self.is_construct_name_field(position):
            return self.target_child(position, symbol)
        return super().on_child_enter(position, symbol)

    def on_child_exit(
        self, position: int, symbol: int, child: Handler
    ) -> Callable[[], None]:
        undos = []
        if self.is_construct_name_field(position):
            print(self, [self.mask.id_to_name(x) for x in child.defined_symbols])
            for idx_list in (
                self.original_defined_production_idxs,
                self.defined_production_idxs,
            ):
                additional_elements = [
                    x for x in child.defined_symbols if x not in idx_list
                ]
                idx_list += additional_elements
                undos.append(remove_last_n_elements(idx_list, len(additional_elements)))
        undos.append(super().on_child_exit(position, symbol, child))
        return chain_undos(undos)

    def is_construct_name_field(self, position):
        return (
            self.construct_name_field is not None
            and position == self.child_fields[self.construct_name_field]
        )

    def is_defining(self, position: int) -> bool:
        if position == self.child_fields[self.construct_name_field]:
            return True
        return super().is_defining(position)


class FuncDefHandler(DefiningConstructHandler):
    name = "FunctionDef~S"
    targeted = ["args"]
    define_symbols_on_exit = "args"
    construct_name_field = "name"


class ClassDefHandler(DefiningConstructHandler):
    name = "ClassDef~S"
    targeted = []
    define_symbols_on_exit = "decorator_list"
    construct_name_field = "name"


defining_construct_handler = [FuncDefHandler, ClassDefHandler]



from typing import Callable, Tuple

from neurosym.program_dist.tree_distribution.preorder_mask.undos import (
    remove_last_n_elements,
)
from neurosym.python_dsl.names import PYTHON_DSL_SEPARATOR

from .handler import ConstructHandler, Handler


def create_target_handler(
    position: int, root_symbol: int, mask, defined_production_idxs, config
):
    """
    Create a target handler for the given root symbol. A target handler collects
        the assigned targets in the given expression.
    """
    targets_map = {
        "Name~L": NameTargetHandler,
        "arg~A": ArgTargetHandler,
        "alias~alias": AliasTargetHandler,
        "const-None~A": NonCollectingTargetHandler,
        "Subscript~L": NonCollectingTargetHandler,
        "Attribute~L": NonCollectingTargetHandler,
        "Tuple~L": TupleLHSHandler,
        "List~L": ListLHSHandler,
        "_starred_content~L": PassthroughLHSHandler,
        "_starred_starred~L": PassthroughLHSHandler,
        "Starred~L": StarredHandler,
        "arguments~As": ArgumentsHandler,
    }

    symbol = root_symbol
    symbol = mask.id_to_name(symbol)

    if symbol.startswith("const") and symbol.split(PYTHON_DSL_SEPARATOR)[-1] in {
        "Name",
        "NullableName",
        "NameStr",
        "NullableNameStr",
    }:
        return SymbolTargetHandler(mask, defined_production_idxs, config, root_symbol)

    pulled = config.pull_handler(
        position,
        symbol,
        mask,
        defined_production_idxs,
        handler_fn=create_target_handler,
    )
    if pulled is not None:
        return pulled

    if symbol.startswith("list"):
        return PassthroughLHSHandler(mask, defined_production_idxs, config)
    return targets_map[symbol](mask, defined_production_idxs, config)


class TargetHandler(Handler):
    """
    Handler that collects targets, stored in defined_symbols.
    """

    def __init__(self, mask, defined_production_idxs, config):
        super().__init__(mask, defined_production_idxs, config)
        self.defined_symbols = []

    def on_child_exit(
        self, position: int, symbol: int, child: Handler
    ) -> Callable[[], None]:
        if hasattr(child, "defined_symbols"):
            symbols = [
                x for x in child.defined_symbols if x not in self.defined_symbols
            ]
            self.defined_symbols += symbols
            return remove_last_n_elements(self.defined_symbols, len(symbols))
        return lambda: None


class TargetConstructHandler(TargetHandler, ConstructHandler):
    def __init__(self, mask, defined_production_idxs, config):
        TargetHandler.__init__(self, mask, defined_production_idxs, config)
        ConstructHandler.__init__(self, mask, defined_production_idxs, config)


class PassthroughLHSHandler(TargetHandler):
    """
    Pass through handler that does not collect any information,
        instead it just targets the children at the given indices.

    If indices is None, it will target all children.
    """

    def on_child_enter(
        self, position: int, symbol: int
    ) -> Tuple[Handler, Callable[[], None]]:
        if self.is_defining(position):
            return self.target_child(position, symbol)
        return super().on_child_enter(position, symbol)

    def is_defining(self, position: int) -> bool:
        return True


class PassthroughLHSConstructHandler(PassthroughLHSHandler, TargetConstructHandler):
    """
    Pass through handler that is also a construct handler.
    """

    use_fields: list[str] = None

    def __init__(self, mask, defined_production_idxs, config):
        PassthroughLHSHandler.__init__(self, mask, defined_production_idxs, config)
        TargetConstructHandler.__init__(self, mask, defined_production_idxs, config)

        self.indices = [self.child_fields[x] for x in self.use_fields]

    def is_defining(self, position: int) -> bool:
        return self.indices is None or position in self.indices


class NonCollectingTargetHandler(PassthroughLHSHandler):
    """
    This is for LHS values where nothing is actually being defined
        (e.g., Subscript, Attribute, etc.)
    """

    def is_defining(self, position: int) -> bool:
        return False


class SymbolTargetHandler(TargetHandler):
    """
    Target handler for symbols. Since symbols have no children, this takes in the symbol
        and sets it up in the defined symbols in __init__.

    Do not keep the symbol if it is a const-None~NullableNameStr.
    """

    def __init__(self, mask, defined_production_idxs, config, symbol):
        super().__init__(mask, defined_production_idxs, config)
        if self.mask.id_to_name(symbol) != "const-None~NullableNameStr":
            self.defined_symbols = [symbol]

    def on_child_enter(
        self, position: int, symbol: int
    ) -> Tuple[Handler, Callable[[], None]]:
        raise NotImplementedError("symbols should not have children.")

    def on_child_exit(
        self, position: int, symbol: int, child: Handler
    ) -> Callable[[], None]:
        raise NotImplementedError("symbols should not have children.")

    def is_defining(self, position: int) -> bool:
        return True


class StarredHandler(PassthroughLHSConstructHandler):
    """
    Handles starred expressions.
    """

    name = "Starred~L"
    use_fields = ["value"]


class ArgumentsHandler(PassthroughLHSConstructHandler):
    """
    Handles arguments.
    """

    name = "arguments~As"
    use_fields = ("posonlyargs", "args", "vararg", "kwonlyargs", "kwarg")


class TupleLHSHandler(PassthroughLHSConstructHandler):
    """
    This is for LHS values where nothing is actually being defined
        (e.g., Subscript, Attribute, etc.)
    """

    name = "Tuple~L"
    use_fields = ["elts"]


class ListLHSHandler(TupleLHSHandler):
    """
    Handler for list LHS values.
    """

    name = "List~L"


class NameTargetHandler(TargetConstructHandler):
    """
    Handler for Name values, as well as other constructs containing symbols.
        Always keeps only the last defined symbol.
    """

    name = "Name~Name"
    # will select the last of these that is defined
    name_nodes = {"id"}

    def on_child_enter(
        self, position: int, symbol: int
    ) -> Tuple[Handler, Callable[[], None]]:
        if self.is_defining(position):
            return self.target_child(position, symbol)
        return super().on_child_enter(position, symbol)

    def on_child_exit(
        self, position: int, symbol: int, child: Handler
    ) -> Callable[[], None]:
        if hasattr(child, "defined_symbols") and child.defined_symbols:
            previous = self.defined_symbols
            self.defined_symbols = child.defined_symbols
            return lambda: setattr(self, "defined_symbols", previous)
        return lambda: None

    def is_defining(self, position: int) -> bool:
        return any(position == self.child_fields[x] for x in self.name_nodes)


class ArgTargetHandler(NameTargetHandler):
    name = "arg~arg"
    name_nodes = {"arg"}


class AliasTargetHandler(NameTargetHandler):
    name = "alias~alias"
    name_nodes = {"name", "asname"}



import copy

from neurosym.programs.s_expression import SExpression
from neurosym.python_dsl.names import PYTHON_DSL_SEPARATOR
from neurosym.python_dsl.python_ast_tools import clean_type, is_sequence


def run_dfa_on_program(dfa, node, state):
    """
    Runs the dfa on a program, yielding the node and the state
        for each node in the program.

    Args:
        dfa: The dfa.
        node: The node.
        state: The root state

    Yields:
        A tuple of the node and the state, for each node in the program.
    """
    if not isinstance(node, (SExpression, str)):
        raise ValueError(f"expected SExpression or str, got {node}")
    yield node, state
    if not isinstance(node, SExpression):
        return
    if not node.children:
        # avoid looking up dfa[state][tag] when there are no children; allows sparser dfa
        return
    if state not in dfa:
        raise ValueError(f"state {state} not in dfa")
    if node.symbol not in dfa[state]:
        raise ValueError(f"symbol {node.symbol} not in dfa[{state}]")
    dfa_states = dfa[state][node.symbol]
    for i, child in enumerate(node.children):
        yield from run_dfa_on_program(dfa, child, dfa_states[i % len(dfa_states)])


def add_disambiguating_type_tags(dfa, prog, start_state):
    """
    Add disambiguating type tags to a program, which appended to each symbol in the program,
        after the separator. Also adds a sequence length tag if the symbol is a sequence type.

    Args:
        dfa: The dfa.
        prog: The program.
        start_state: The state for the root node.

    Returns:
        The program with disambiguating type tags.
    """
    prog = copy.deepcopy(prog)
    node_id_to_new_symbol = {}
    for node, tag in run_dfa_on_program(dfa, prog, start_state):
        assert isinstance(node, SExpression), node
        new_symbol = node.symbol + PYTHON_DSL_SEPARATOR + clean_type(tag)
        if is_sequence(tag, node.symbol):
            new_symbol += PYTHON_DSL_SEPARATOR + str(len(node.children))
        node_id_to_new_symbol[id(node)] = new_symbol
    return prog.replace_symbols_by_id(node_id_to_new_symbol)



from types import NoneType
from typing import Callable, Tuple, Union

import numpy as np

from neurosym.program_dist.tree_distribution.preorder_mask.preorder_mask import (
    PreorderMask,
)
from neurosym.program_dist.tree_distribution.tree_distribution import TreeDistribution
from neurosym.programs.s_expression import SExpression


def compute_likelihood(
    tree_dist: TreeDistribution,
    program: SExpression,
    parents: Tuple[Tuple[int, int], ...],
    preorder_mask: PreorderMask,
    tracker: Union[NoneType, Callable[[SExpression, float], NoneType]],
):
    """
    Compute the likelihood of a program under a distribution.

    If `tracker` is not None, it will be called with each node and the likelihood
        of that node. This can be useful for debugging why a program has a certain
        likelihood.
    """
    start_position = parents[-1][1]
    top_symbol = tree_dist.symbol_to_index[program.symbol]
    likelihood = symbol_likelihood(
        tree_dist, parents, preorder_mask, start_position, top_symbol
    )
    if tracker is not None:
        tracker(program, likelihood)
    # only end early if the tracker is None,
    # because we want to call the tracker otherwise
    elif likelihood == -float("inf"):
        return -float("inf")
    preorder_mask.on_entry(start_position, top_symbol)
    order = tree_dist.ordering.order(top_symbol, len(program.children))
    for i, child in zip(order, [program.children[i] for i in order]):
        likelihood += compute_likelihood(
            tree_dist,
            child,
            (parents + ((top_symbol, i),))[-tree_dist.limit :],
            preorder_mask=preorder_mask,
            tracker=tracker,
        )
        if likelihood == -float("inf") and tracker is None:
            return -float("inf")
    preorder_mask.on_exit(start_position, top_symbol)
    return likelihood


def symbol_likelihood(tree_dist, parents, preorder_mask, start_position, top_symbol):
    if parents not in tree_dist.likelihood_arrays:
        return -float("inf")
    idx = tree_dist.index_within_distribution_list[parents].get(top_symbol, None)
    if idx is None:
        return -float("inf")
    syms, log_probs = tree_dist.likelihood_arrays[parents]
    mask = preorder_mask.compute_mask(start_position, syms)
    if not mask[idx]:
        return -float("inf")
    denominator = np.logaddexp.reduce(log_probs[mask])
    likelihood = log_probs[idx] - denominator

    assert likelihood <= 0, f"Likelihood is {likelihood}, expected <= 0"

    return likelihood



from dataclasses import dataclass
from typing import Dict, Tuple


@dataclass(frozen=True, eq=True)
class SExpression:
    symbol: str
    children: Tuple["SExpression"]

    @property
    def postorder(self):
        for x in self.children:
            yield from x.postorder
        yield self

    def replace_symbols_by_id(self, id_to_symbol):
        """
        Replace symbols in the S-expression based on the given mapping.

        Args:
            id_to_symbol: A mapping from id(node) to a new symbol.
        """
        return SExpression(
            id_to_symbol.get(id(self), self.symbol),
            tuple(child.replace_symbols_by_id(id_to_symbol) for child in self.children),
        )

    def replace_nodes_by_id(self, id_to_new_node):
        """
        Replace nodes in the S-expression based on the given mapping. Any remapped
            node will not have its children replaced.

        Args:
            id_to_new_node: A mapping from id(node) to a new node.
        """
        if id(self) in id_to_new_node:
            return id_to_new_node[id(self)]
        return SExpression(
            self.symbol,
            tuple(child.replace_nodes_by_id(id_to_new_node) for child in self.children),
        )


@dataclass
class InitializedSExpression:
    symbol: str
    children: Tuple["InitializedSExpression"]
    # state includes things related to the execution of the program,
    # e.g. weights of a neural network
    state: Dict[str, object]

    def all_state_values(self):
        yield from self.state.values()
        for child in self.children:
            yield from child.all_state_values()



import ast
import re
import unittest
from textwrap import dedent

from increase_recursionlimit import increase_recursionlimit
from parameterized import parameterized

import neurosym as ns
from tests.python.parse_test import small_set_examples

dfa = ns.python_dfa()

reasonable_classifications = [
    ("alias", "alias"),
    ("AnnAssign", "S"),
    ("Assert", "S"),
    ("Assign", "S"),
    ("AsyncFunctionDef", "S"),
    ("AsyncFor", "S"),
    ("AsyncWith", "S"),
    ("Attribute", "E"),
    ("Attribute", "L"),
    ("AugAssign", "S"),
    ("Await", "E"),
    ("BinOp", "E"),
    ("BoolOp", "E"),
    ("Call", "E"),
    ("ClassDef", "S"),
    ("Compare", "E"),
    ("Constant", "E"),
    ("Constant", "F"),
    ("Delete", "S"),
    ("Dict", "E"),
    ("DictComp", "E"),
    *[
        (x, "O")
        for x in [
            "Add",
            "Sub",
            "Mult",
            "MatMult",
            "Div",
            "Mod",
            "Pow",
            "LShift",
            "RShift",
            "BitOr",
            "BitXor",
            "BitAnd",
            "FloorDiv",
            "Invert",
            "Not",
            "UAdd",
            "USub",
            "Eq",
            "NotEq",
            "Lt",
            "LtE",
            "Gt",
            "GtE",
            "Is",
            "IsNot",
            "In",
            "NotIn",
            "And",
            "Or",
        ]
    ],
    ("ExceptHandler", "EH"),
    ("Expr", "S"),
    ("For", "S"),
    ("FormattedValue", "F"),
    ("FunctionDef", "S"),
    ("GeneratorExp", "E"),
    ("Global", "S"),
    ("If", "S"),
    ("IfExp", "E"),
    ("Import", "S"),
    ("ImportFrom", "S"),
    ("JoinedStr", "E"),
    ("JoinedStr", "F"),
    ("Lambda", "E"),
    ("List", "E"),
    ("List", "L"),
    ("ListComp", "E"),
    ("Module", "M"),
    ("Name", "E"),
    ("Name", "L"),
    ("NamedExpr", "E"),
    ("Nonlocal", "S"),
    ("Pass", "S"),
    ("Raise", "S"),
    ("Return", "S"),
    ("Break", "S"),
    ("Continue", "S"),
    ("Set", "E"),
    ("SetComp", "E"),
    ("_slice_content", "SliceRoot"),
    ("_slice_slice", "SliceRoot"),
    ("_slice_tuple", "SliceRoot"),
    ("Tuple", "SliceTuple"),
    ("Slice", "Slice"),
    ("Starred", "L"),
    ("Starred", "Starred"),
    ("_starred_content", "StarredRoot"),
    ("_starred_content", "L"),
    ("_starred_starred", "StarredRoot"),
    ("_starred_starred", "L"),
    ("Subscript", "E"),
    ("Subscript", "L"),
    ("Try", "S"),
    ("Tuple", "E"),
    ("Tuple", "L"),
    ("UnaryOp", "E"),
    ("While", "S"),
    ("With", "S"),
    ("Yield", "E"),
    ("YieldFrom", "E"),
    ("arg", "A"),
    ("arguments", "As"),
    ("comprehension", "C"),
    ("keyword", "K"),
    ("list", "[A]"),
    ("list", "[C]"),
    ("list", "[SliceRoot]"),
    ("list", "[E]"),
    ("list", "[StarredRoot]"),
    ("list", "[EH]"),
    ("list", "[F]"),
    ("list", "[K]"),
    ("list", "[L]"),
    ("list", "[W]"),
    ("list", "[O]"),
    ("list", "[alias]"),
    ("list", "[NameStr]"),
    ("list", "[TI]"),
    ("/seq", "seqS"),
    ("withitem", "W"),
    ("const-None", "E"),
    # formatting
    ("const-None", "F"),
    ("const-s.*", "F"),
    # vararg
    ("const-None", "A"),
    # type
    (".*", "TA"),
    ("const-None", "TC"),
    # left value
    ("const-None", "L"),
    # Load/Store/Del
    ("Load", "Ctx"),
    ("Store", "Ctx"),
    ("Del", "Ctx"),
    # name
    ("const-[&g].*", "Name"),
    ("const-[&g].*", "NullableName"),
    ("const-None", "NullableName"),
    ("const-s.*", "NameStr"),
    ("const-[&g].*", "NameStr"),  # imports
    ("const-s.*", "NullableNameStr"),
    ("const-None", "NullableNameStr"),
    ("const-[&g].*", "NullableNameStr"),
    # values
    ("const-None", "Const"),
    ("const-True", "Const"),
    ("const-False", "Const"),
    ("const-Ellipsis", "Const"),
    ("const-[sbijf].*", "Const"),
    # constkind
    ("const-None", "ConstKind"),
    ("const-s.*", "ConstKind"),
    # constants
    ("const-i[01]", "bool"),
    ("const-i.*", "int"),
]


class TestClassifications(unittest.TestCase):
    def classify_in_code(self, code, start_state):
        classified = [
            (ns.render_s_expression(x), tag)
            for x, tag in ns.run_dfa_on_program(
                dfa, code.to_ns_s_exp(dict()), start_state
            )
            if isinstance(x, ns.SExpression)
        ]
        print(classified)
        return classified

    def test_module_classify(self):
        self.assertEqual(
            self.classify_in_code(ns.python_to_python_ast("x = 2"), "M"),
            [
                (
                    "(Module (/seq (Assign (list (Name &x:0 Store)) (Constant i2 None) None)) nil)",
                    "M",
                ),
                (
                    "(/seq (Assign (list (Name &x:0 Store)) (Constant i2 None) None))",
                    "seqS",
                ),
                ("(Assign (list (Name &x:0 Store)) (Constant i2 None) None)", "S"),
                ("(list (Name &x:0 Store))", "[L]"),
                ("(Name &x:0 Store)", "L"),
                ("(Constant i2 None)", "E"),
            ],
        )

    def test_statement_classify(self):
        self.assertEqual(
            self.classify_in_code(ns.python_statement_to_python_ast("x = 2"), "S"),
            [
                (
                    "(Assign (list (Name &x:0 Store)) (Constant i2 None) None)",
                    "S",
                ),
                ("(list (Name &x:0 Store))", "[L]"),
                ("(Name &x:0 Store)", "L"),
                ("(Constant i2 None)", "E"),
            ],
        )


class DFATest(unittest.TestCase):
    def check_reasonable_classification(self, tag_to_check, state_to_check):
        for tag, state in reasonable_classifications:
            mat = re.match("^" + tag + "$", tag_to_check)
            if mat and state == state_to_check:
                return
        self.fail(f"Unknown classification {tag_to_check} {state_to_check}")

    def classify_elements_in_code_with_config(self, code, **kwargs):
        with increase_recursionlimit():
            print("#" * 80)
            print(code)
            code = ns.python_to_python_ast(code).to_ns_s_exp(kwargs)
            print(ns.render_s_expression(code))
            classified = ns.run_dfa_on_program(dfa, code, "M")
            result = sorted(
                {
                    (x.symbol, state)
                    for (x, state) in classified
                    if isinstance(x, ns.SExpression)
                }
            )
            print(code)
            for x, state in result:
                self.check_reasonable_classification(x, state)

    def classify_elements_in_code(self, code):
        self.classify_elements_in_code_with_config(code)
        self.classify_elements_in_code_with_config(code, no_leaves=True)

    @parameterized.expand([(i,) for i in range(100)])
    def test_realistic(self, i):
        code = small_set_examples()[i]
        for element in ast.parse(code).body:
            self.classify_elements_in_code(ast.unparse(element))

    def test_function(self):
        self.classify_elements_in_code(
            dedent(
                """
                def f(x):
                    pass
                """
            )
        )
        self.classify_elements_in_code(
            dedent(
                """
                def f(x, *y, z=2, **w):
                    pass
                """
            )
        )

    def test_with(self):
        self.classify_elements_in_code(
            dedent(
                """
                with x:
                    pass
                """
            )
        )
        self.classify_elements_in_code(
            dedent(
                """
                with x as y:
                    pass
                """
            )
        )

    def test_annotation(self):
        self.classify_elements_in_code(
            dedent(
                """
                @asyncio.coroutine
                def onOpen(self):
                    pass
                """
            )
        )

    def test_class(self):
        self.classify_elements_in_code(
            dedent(
                """
                class MyClientProtocol(WebSocketClientProtocol):
                    pass
                """
            )
        )

    def test_exception(self):
        self.classify_elements_in_code(
            dedent(
                """
                try:
                    x = 2
                except:
                    pass
                """
            )
        )

    def test_comparison(self):
        self.classify_elements_in_code("x == 2")

    def test_keyword(self):
        self.classify_elements_in_code("f(x=2)")

    def test_aug_assign(self):
        self.classify_elements_in_code("(x := 2)")

    def test_assign(self):
        self.classify_elements_in_code("x = 2")
        self.classify_elements_in_code("x = 2, 3")
        self.classify_elements_in_code("x += 2")

    def test_tuple(self):
        self.classify_elements_in_code("(2, 3)")

    def test_unpacking(self):
        self.classify_elements_in_code("a, b = 1, 2")
        self.classify_elements_in_code("a, *b = 1, 2, 3")
        self.classify_elements_in_code("[a, b] = 1, 2, 3")

    def test_slicing_direct(self):
        self.classify_elements_in_code("x[2]")
        self.classify_elements_in_code("x[2:3]")
        self.classify_elements_in_code("x[2:3] = 4")

    def test_slicing_tuple(self):
        self.classify_elements_in_code("x[2:3, 3]")
        self.classify_elements_in_code("x[2:3, 3] = 5")

    def test_starred(self):
        self.classify_elements_in_code("f(2, 3, *x)")
        self.classify_elements_in_code("(2, 3, *x)")
        self.classify_elements_in_code("[2, 3, *x]")
        self.classify_elements_in_code("{2, 3, *x}")

    def test_kwstarred(self):
        self.classify_elements_in_code("f(2, 3, **x)")

    def test_import(self):
        self.classify_elements_in_code("import x")
        self.classify_elements_in_code("from x import y")
        self.classify_elements_in_code("from x import y as z")
        self.classify_elements_in_code("from . import x")

    def test_global_nonlocal(self):
        self.classify_elements_in_code("global x")
        self.classify_elements_in_code("nonlocal x")

    def test_joined_str(self):
        self.classify_elements_in_code("f'2 {459.67:.1f}'")

    def test_type_annotation(self):
        self.classify_elements_in_code("x: int = 2")
        self.classify_elements_in_code("def f(x: int) -> int: pass")
        self.classify_elements_in_code("x: List[int] = []")
        self.classify_elements_in_code("x: Tuple[int, int] = (x, y)")

    def test_code(self):
        self.classify_elements_in_code(
            dedent(
                r"""
                {x: 2 for x in range(10)}
                """
            )
        )


class TestExprNodeValidity(unittest.TestCase):
    def e_nodes(self, code):
        with increase_recursionlimit():
            print("#" * 80)
            print(code)
            code = ns.python_to_python_ast(code)
            e_nodes = [
                ns.render_s_expression(x)
                for x, state in ns.run_dfa_on_program(
                    dfa, code.to_ns_s_exp(dict()), "M"
                )
                if state == "E" and isinstance(x, ns.SExpression)
            ]
            return e_nodes

    def assertENodeReal(self, node):
        print(node)
        code = ns.s_exp_to_python_ast(node)
        print(code)
        code_in_function_call = ns.make_python_ast.make_call(
            ns.PythonSymbol(name="hi", scope=None), code
        )
        code_in_function_call = code_in_function_call.to_python()
        print(code_in_function_call)
        code_in_function_call = ns.python_statement_to_python_ast(code_in_function_call)
        assert code_in_function_call.typ == ast.Expr
        code_in_function_call = code_in_function_call.children[0]
        assert code_in_function_call.typ == ast.Call
        code_in_function_call = code_in_function_call.children[1]
        code_in_function_call = code_in_function_call.children[0].content
        print(code_in_function_call)
        code_in_function_call = code_in_function_call.to_python()
        print(code_in_function_call)
        self.maxDiff = None
        self.assertEqual(code.to_python(), code_in_function_call)

    def assertENodesReal(self, code):
        e_nodes = self.e_nodes(code)
        for node in e_nodes:
            self.assertENodeReal(node)

    def test_e_nodes_basic(self):
        e_nodes = self.e_nodes("x == 2")
        self.assertEqual(
            e_nodes,
            [
                "(Compare (Name g_x Load) (list Eq) (list (Constant i2 None)))",
                "(Name g_x Load)",
                "(Constant i2 None)",
            ],
        )

    def test_slice(self):
        self.assertENodesReal("y = x[2:3]")

    @parameterized.expand([(i,) for i in range(100)])
    def test_realistic(self, i):
        code = small_set_examples()[i]
        for element in ast.parse(code).body:
            self.assertENodesReal(ast.unparse(element))



import itertools
import unittest

import neurosym as ns


def make_compute_dsl():
    dslf = ns.DSLFactory()
    dslf.concrete("+", "(i, i) -> i", lambda x, y: x + y)
    dslf.concrete("1", "() -> i", lambda: 1)
    dslf.concrete("double", "(i) -> i", lambda x: x * 2)
    dslf.concrete("triple", "(i) -> i", lambda x: x * 3)
    dslf.lambdas()
    dslf.prune_to("i", "i -> i", "i -> i -> i", "(i, i) -> i", "(i, i) -> i -> i")
    return dslf.finalize()


compute_dsl = make_compute_dsl()


class TestEvaluate(unittest.TestCase):
    def test_show_dsl(self):
        expected = """
            + :: (i, i) -> i
            1 :: () -> i
        double :: i -> i
        triple :: i -> i
        lam_0 :: L<#body|i;i> -> (i, i) -> #body
        lam_1 :: L<#body|i> -> i -> #body
        $0_0 :: V<i@0>
        $1_0 :: V<i@1>
        $2_0 :: V<i@2>
        """
        actual = compute_dsl.render()
        self.assertEqual(
            {line.strip() for line in actual.strip().split("\n")},
            {line.strip() for line in expected.strip().split("\n")},
        )

    def evaluate(self, code):
        return compute_dsl.compute(compute_dsl.initialize(ns.parse_s_expression(code)))

    def test_constant(self):
        fn = self.evaluate("(lam_1 (1))")
        for i in range(10):
            self.assertEqual(fn(i), 1)

    def test_identity(self):
        fn = self.evaluate("(lam_1 ($0_0))")
        for i in range(10):
            self.assertEqual(fn(i), i)

    def test_add_1(self):
        fn = self.evaluate("(lam_1 (+ (1) ($0_0)))")
        for i in range(10):
            self.assertEqual(fn(i), i + 1)

    def test_double_constant(self):
        fn = self.evaluate("(lam_1 (lam_1 (1)))")
        for i in range(10):
            for j in range(10):
                self.assertEqual(fn(i)(j), 1)

    def test_get_inner_var(self):
        fn = self.evaluate("(lam_1 (lam_1 ($0_0)))")
        for i in range(10):
            for j in range(10):
                self.assertEqual(fn(i)(j), j)

    def test_get_outer_var(self):
        fn = self.evaluate("(lam_1 (lam_1 ($1_0)))")
        for i in range(10):
            for j in range(10):
                self.assertEqual(fn(i)(j), i)

    def test_sub_curried(self):
        fn = self.evaluate("(lam_1 (lam_1 (+ (double ($0_0)) ($1_0))))")
        for i in range(10):
            for j in range(10):
                self.assertEqual(fn(i)(j), i + 2 * j)

    def test_two_arg_const(self):
        fn = self.evaluate("(lam_0 (1))")
        for i in range(10):
            for j in range(10):
                self.assertEqual(fn(i, j), 1)

    def test_two_arg_sub(self):
        fn = self.evaluate("(lam_0 (+ (double ($0_0)) ($1_0)))")
        for i in range(10):
            for j in range(10):
                self.assertEqual(fn(i, j), j * 2 + i)

    def test_three_arg_sub(self):
        fn = self.evaluate(
            """
            (lam_0 (lam_1
                (+
                    (+
                        (triple ($0_0))
                        (double ($1_0)))
                    ($2_0))))
            """
        )
        for i in range(10):
            for j in range(10):
                for k in range(10):
                    self.assertEqual(fn(i, j)(k), k * 3 + j * 2 + i)


class TestEnumerateBasicArithmetic(unittest.TestCase):
    def test_show_dsl(self):
        self.maxDiff = None
        expected = """
            + :: (i, i) -> i
            1 :: () -> i
        lam_0 :: L<#body|i;i> -> (i, i) -> #body
        lam_1 :: L<#body|i> -> i -> #body
        $0_0 :: V<i@0>
        $1_0 :: V<i@1>
        $2_0 :: V<i@2>
        """
        actual = ns.examples.basic_arith_dsl(True).render()
        self.assertEqual(
            {line.strip() for line in actual.strip().split("\n")},
            {line.strip() for line in expected.strip().split("\n")},
        )

    def assertEnumerateType(self, typ, expected, filt=lambda x: True):
        g = ns.DSLSearchGraph(
            ns.examples.basic_arith_dsl(True),
            ns.parse_type(typ),
            ns.ChooseFirst(),
            filt,
            metadata_computer=ns.NoMetadataComputer(),
        )

        res = [
            ns.render_s_expression(prog.program)
            for prog in itertools.islice(ns.search.bfs(g, 1000), 10)
        ]

        print(res)

        self.assertEqual(res, expected)

    def test_single(self):
        self.assertEnumerateType(
            "i -> i",
            [
                "(lam_1 (1))",
                "(lam_1 ($0_0))",
                "(lam_1 (+ (1) (1)))",
                "(lam_1 (+ (1) ($0_0)))",
                "(lam_1 (+ ($0_0) (1)))",
                "(lam_1 (+ ($0_0) ($0_0)))",
                "(lam_1 (+ (+ (1) (1)) (1)))",
                "(lam_1 (+ (+ (1) (1)) ($0_0)))",
                "(lam_1 (+ (+ (1) ($0_0)) (1)))",
                "(lam_1 (+ (+ (1) ($0_0)) ($0_0)))",
            ],
        )

    def test_curried(self):
        self.assertEnumerateType(
            "i -> i -> i",
            [
                "(lam_1 (lam_1 (1)))",
                "(lam_1 (lam_1 ($0_0)))",
                "(lam_1 (lam_1 ($1_0)))",
                "(lam_1 (lam_1 (+ (1) (1))))",
                "(lam_1 (lam_1 (+ (1) ($0_0))))",
                "(lam_1 (lam_1 (+ (1) ($1_0))))",
                "(lam_1 (lam_1 (+ ($0_0) (1))))",
                "(lam_1 (lam_1 (+ ($0_0) ($0_0))))",
                "(lam_1 (lam_1 (+ ($0_0) ($1_0))))",
                "(lam_1 (lam_1 (+ ($1_0) (1))))",
            ],
        )

    def test_2arg(self):
        self.assertEnumerateType(
            "(i, i) -> i",
            [
                "(lam_0 (1))",
                "(lam_0 ($0_0))",
                "(lam_0 ($1_0))",
                "(lam_0 (+ (1) (1)))",
                "(lam_0 (+ (1) ($0_0)))",
                "(lam_0 (+ (1) ($1_0)))",
                "(lam_0 (+ ($0_0) (1)))",
                "(lam_0 (+ ($0_0) ($0_0)))",
                "(lam_0 (+ ($0_0) ($1_0)))",
                "(lam_0 (+ ($1_0) (1)))",
            ],
        )

    def test_3arg_mix_curried(self):
        self.assertEnumerateType(
            "(i, i) -> i -> i",
            [
                "(lam_0 (lam_1 ($0_0)))",
                "(lam_0 (lam_1 ($1_0)))",
                "(lam_0 (lam_1 ($2_0)))",
                "(lam_0 (lam_1 (+ (1) ($0_0))))",
                "(lam_0 (lam_1 (+ (1) ($1_0))))",
                "(lam_0 (lam_1 (+ (1) ($2_0))))",
                "(lam_0 (lam_1 (+ ($0_0) (1))))",
                "(lam_0 (lam_1 (+ ($0_0) ($0_0))))",
                "(lam_0 (lam_1 (+ ($0_0) ($1_0))))",
                "(lam_0 (lam_1 (+ ($0_0) ($2_0))))",
            ],
            filt=lambda x: "$" in str(x.program),
        )


def make_varied_type_dsl():
    dslf = ns.DSLFactory()
    dslf.concrete("^", "(f, i) -> f", lambda x, y: x**y)
    dslf.concrete("1", "() -> i", lambda: 1)
    dslf.concrete("1f", "() -> f", lambda x: 1.0)
    dslf.concrete("double", "(i) -> i", lambda x: x * 2)
    dslf.lambdas()
    dslf.prune_to("f", "f -> f", "(f, i) -> f", "(f, i) -> i")
    return dslf.finalize()


class TestVariedTypes(unittest.TestCase):
    def test_show_dsl(self):
        self.maxDiff = None
        expected = """
              ^ :: (f, i) -> f
              1 :: () -> i
             1f :: () -> f
         double :: i -> i
          lam_0 :: L<#body|f;i> -> (f, i) -> #body
          lam_1 :: L<#body|f> -> f -> #body
           $0_0 :: V<f@0>
           $1_0 :: V<f@1>
           $0_1 :: V<i@0>
        """
        actual = make_varied_type_dsl().render()
        print(actual)
        self.assertEqual(
            {line.strip() for line in actual.strip().split("\n")},
            {line.strip() for line in expected.strip().split("\n")},
        )

    def assertEnumerateType(self, typ, expected, filt=lambda x: True):
        g = ns.DSLSearchGraph(
            make_varied_type_dsl(),
            ns.parse_type(typ),
            ns.ChooseFirst(),
            filt,
            metadata_computer=ns.NoMetadataComputer(),
        )

        res = [
            ns.render_s_expression(prog.program)
            for prog in itertools.islice(ns.search.bfs(g, 1000), 10)
        ]

        print(res)

        self.assertEqual(res, expected)

    def test_fi(self):
        self.assertEnumerateType(
            "(f, i) -> f",
            [
                "(lam_0 (1f))",
                "(lam_0 ($1_0))",
                "(lam_0 (^ (1f) (1)))",
                "(lam_0 (^ (1f) ($0_1)))",
                "(lam_0 (^ ($1_0) (1)))",
                "(lam_0 (^ ($1_0) ($0_1)))",
                "(lam_0 (^ (1f) (double (1))))",
                "(lam_0 (^ (1f) (double ($0_1))))",
                "(lam_0 (^ ($1_0) (double (1))))",
                "(lam_0 (^ ($1_0) (double ($0_1))))",
            ],
        )



import unittest

import neurosym as ns


class DSLSubsetTest(unittest.TestCase):
    def setUp(self):
        self.dfa = ns.python_dfa()

    def test_subset_basic(self):
        subset = ns.PythonDSLSubset.from_programs(
            self.dfa,
            ns.python_to_python_ast("x = x + 2; y = y + x + 2"),
            root="M",
        )
        print(subset)
        self.assertEqual(
            subset,
            ns.PythonDSLSubset(
                _lengths_by_sequence_type={"seqS": {2}, "[L]": {1}, "[TI]": {0}},
                _leaves={
                    "Name": {"const-&x:0", "const-&y:0"},
                    "Ctx": {"Load", "Store"},
                    "O": {"Add"},
                    "Const": {"const-i2"},
                    "ConstKind": {"const-None"},
                    "TC": {"const-None"},
                },
            ),
        )

    def test_subset_multi_length(self):
        subset = ns.PythonDSLSubset.from_programs(
            self.dfa,
            ns.python_to_python_ast("x = [1, 2, 3]; y = [1, 2, 3, 4]"),
            root="M",
        )
        print(subset)
        self.assertEqual(
            subset,
            ns.PythonDSLSubset(
                _lengths_by_sequence_type={
                    "seqS": {2},
                    "[L]": {1},
                    "[StarredRoot]": {3, 4},
                    "[TI]": {0},
                },
                _leaves={
                    "Name": {"const-&x:0", "const-&y:0"},
                    "Ctx": {"Load", "Store"},
                    "Const": {"const-i1", "const-i2", "const-i3", "const-i4"},
                    "ConstKind": {"const-None"},
                    "TC": {"const-None"},
                },
            ),
        )

    def test_subset_multi_root(self):
        subset = ns.PythonDSLSubset.from_programs(
            self.dfa,
            ns.python_to_python_ast("x = x + 2; y = y + x + 2"),
            ns.python_statement_to_python_ast("while True: pass"),
            root=("M", "S"),
        )
        print(subset)
        self.assertEqual(
            subset,
            ns.PythonDSLSubset(
                _lengths_by_sequence_type={"seqS": {0, 1, 2}, "[L]": {1}, "[TI]": {0}},
                _leaves={
                    "Name": {"const-&x:0", "const-&y:0"},
                    "Ctx": {"Load", "Store"},
                    "O": {"Add"},
                    "Const": {"const-True", "const-i2"},
                    "ConstKind": {"const-None"},
                    "TC": {"const-None"},
                    "S": {"Pass"},
                },
            ),
        )

    def test_subset_fill_in_missing(self):
        subset = ns.PythonDSLSubset.from_programs(
            self.dfa,
            ns.python_to_python_ast("x = x + 2; y = y + x + 2"),
            ns.python_to_python_ast("x = 2; y = 3; z = 4; a = 7"),
            root=("M", "M"),
        )
        print(subset)
        self.assertEqual(
            subset.lengths_by_sequence_type, {"seqS": [2, 4], "[L]": [1], "[TI]": [0]}
        )
        subset.fill_in_missing_lengths()
        print(subset)
        self.assertEqual(
            subset.lengths_by_sequence_type,
            {"seqS": [2, 3, 4], "[L]": [1], "[TI]": [0]},
        )



import unittest

import neurosym as ns


class TestBasicAnnotation(unittest.TestCase):
    def assertBasicAnnotation(self, dfa, program, state, expected):
        result = ns.run_dfa_on_program(dfa, ns.parse_s_expression(program), state)
        self.assertEqual(
            [(ns.render_s_expression(node), state) for node, state in result], expected
        )

    def assertDisambiguatingTypeTags(self, dfa, prog, start_state, expected):
        result = ns.add_disambiguating_type_tags(
            dfa, ns.parse_s_expression(prog), start_state
        )
        self.assertEqual(ns.render_s_expression(result), expected)

    def test_basic(self):
        dfa = {
            "start": {
                "a": ["state1", "state2"],
            },
            "state1": {
                "b": ["state2"],
            },
            "state2": {
                "c": [],
            },
        }
        self.assertBasicAnnotation(
            dfa,
            "(a (b (c)) (c))",
            "start",
            [
                ("(a (b (c)) (c))", "start"),
                ("(b (c))", "state1"),
                ("(c)", "state2"),
                ("(c)", "state2"),
            ],
        )
        self.assertDisambiguatingTypeTags(
            dfa,
            "(a (b (c)) (c))",
            "start",
            "(a~start (b~state1 (c~state2)) (c~state2))",
        )

    def test_repeated_state(self):
        dfa = {
            "start": {
                "a": ["[state1]", "state1"],
            },
            "[state1]": {
                "list": ["state1"],
            },
            "state1": {
                "b": [],
            },
        }
        self.assertBasicAnnotation(
            dfa,
            "(a (list (b) (b) (b)))",
            "start",
            [
                ("(a (list (b) (b) (b)))", "start"),
                ("(list (b) (b) (b))", "[state1]"),
                ("(b)", "state1"),
                ("(b)", "state1"),
                ("(b)", "state1"),
            ],
        )
        self.assertDisambiguatingTypeTags(
            dfa,
            "(a (list (b) (b) (b)))",
            "start",
            "(a~start (list~_state1_~3 (b~state1) (b~state1) (b~state1)))",
        )



from abc import ABC, abstractmethod


class SearchGraph(ABC):
    """
    Where nodes are hashable but can be ids etc.
    """

    @abstractmethod
    def initial_node(self):
        pass

    @abstractmethod
    def expand_node(self, node):
        pass

    @abstractmethod
    def is_goal_node(self, node):
        pass



import unittest

import neurosym as ns


class TestParseRender(unittest.TestCase):
    def assertParseRender(self, s_exp, text, **kwargs):
        self.assertEqual(ns.render_s_expression(s_exp, **kwargs), text)
        self.assertEqual(ns.parse_s_expression(text, **kwargs), s_exp)

    def test_basic_render(self):
        self.assertParseRender(
            ns.SExpression("foo", (ns.SExpression("bar", ()),)),
            "(foo (bar))",
        )
        self.assertParseRender(
            ns.SExpression("foo", (ns.SExpression("bar", ()),)),
            "(foo leaf-bar)",
            for_stitch=True,
        )

    def test_variable_render(self):
        self.assertParseRender(
            ns.SExpression("foo", (ns.SExpression("$0", ()),)),
            "(foo ($0))",
        )
        self.assertParseRender(
            ns.SExpression("foo", (ns.SExpression("$0", ()),)),
            "(foo $0)",
            for_stitch=True,
        )



import queue
from dataclasses import dataclass, field
from typing import Callable

from neurosym.programs.s_expression import SExpression
from neurosym.search_graph.search_graph import SearchGraph


def astar(g: SearchGraph, cost_plus_heuristic: Callable[[SExpression], float]):
    """
    Performs a breadth-first search on the given search graph, yielding each node in the
    order it was visited.

    :param g: Search graph to search over
    :param cost_plus_heuristic: Cost plus heuristic function to use for A*.
        The heuristic function should be admissible, i.e. it should never overestimate
        the cost to reach the goal.
    """
    visited = set()
    fringe = queue.PriorityQueue()

    def add_to_fringe(node):
        fringe.put(AStarNode(cost_plus_heuristic(node), node))

    add_to_fringe(g.initial_node())
    # this is similar to the BFS algorithm
    # pylint: disable=duplicate-code
    while not fringe.empty():
        node = fringe.get().node
        if node in visited:
            continue
        visited.add(node)
        if g.is_goal_node(node):
            yield node
        for child in g.expand_node(node):
            add_to_fringe(child)


@dataclass(order=True)
class AStarNode:
    """
    Represents a node in the A* search tree.
    """

    cost: float
    node: SExpression = field(compare=False)



import unittest

import neurosym as ns

from .bigram_test import fam_with_vars


class CollectPreorderSymbolsTest(unittest.TestCase):
    def assertCollectPreorder(self, expected, code, tree_dist, **kwargs):
        result = ns.collect_preorder_symbols(
            ns.parse_s_expression(code), tree_dist, **kwargs
        )
        result = [
            (ns.render_s_expression(s_exp), [tree_dist.symbols[idx][0] for idx in alts])
            for s_exp, alts, _ in result
        ]

        print(result)
        self.assertEqual(result, expected)

    def test_collect_preorder_symbols(self):
        self.assertCollectPreorder(
            [
                ("(+ (1) (call (lam ($0_0)) (1)))", ["+", "1", "2", "call"]),
                ("(1)", ["+", "1", "2", "call"]),
                ("(call (lam ($0_0)) (1))", ["+", "1", "2", "call"]),
                ("(lam ($0_0))", ["lam"]),
                ("($0_0)", ["$0_0", "+", "1", "2", "call"]),
                ("(1)", ["+", "1", "2", "call"]),
            ],
            "(+ (1) (call (lam ($0_0)) (1)))",
            fam_with_vars.tree_distribution_skeleton,
        )

    def test_collect_preorder_symbols_replace_node(self):
        self.assertCollectPreorder(
            [
                ("(call (lam (+ (1) (1))) (1))", ["+", "1", "2", "call"]),
                ("(lam (+ (1) (1)))", ["lam"]),
                # note that it's now (+ ($0_0) ($0_0)) instead of (+ (1) (1))
                ("(+ ($0_0) ($0_0))", ["$0_0", "+", "1", "2", "call"]),
                ("($0_0)", ["$0_0", "+", "1", "2", "call"]),
                ("($0_0)", ["$0_0", "+", "1", "2", "call"]),
                ("(1)", ["+", "1", "2", "call"]),
            ],
            "(call (lam (+ (1) (1))) (1))",
            fam_with_vars.tree_distribution_skeleton,
            replace_node_midstream=lambda sexp, _1, _2, _3: (
                ns.parse_s_expression("(+ ($0_0) ($0_0))")
                if sexp.symbol == "+"
                else sexp
            ),
        )



from typing import Callable, Iterator, List, Tuple

import numpy as np

from neurosym.program_dist.tree_distribution.preorder_mask.preorder_mask import (
    PreorderMask,
)
from neurosym.program_dist.tree_distribution.tree_distribution import TreeDistribution
from neurosym.programs.s_expression import SExpression


def collect_preorder_symbols(
    s_exp: SExpression,
    tree_dist: TreeDistribution,
    replace_node_midstream: Callable[
        [SExpression, PreorderMask, int, List[int]], SExpression
    ] = None,
    symbol_to_index_fn=None,
) -> Iterator[Tuple[SExpression, List[str], PreorderMask]]:
    """
    Collects the alernate symbols that could have been selected in the tree distribution.
    """
    mask = tree_dist.mask_constructor(tree_dist)
    mask.on_entry(0, 0)
    yield from collect_preorder_symbols_dfs(
        s_exp,
        tree_dist,
        mask,
        ((0, 0),),
        replace_node_midstream=replace_node_midstream,
        symbol_to_index_fn=symbol_to_index_fn,
    )


def collect_preorder_symbols_dfs(
    s_exp: SExpression,
    tree_dist: TreeDistribution,
    mask: PreorderMask,
    parents: Tuple[Tuple[int, int], ...],
    replace_node_midstream: Callable[
        [SExpression, PreorderMask, int, List[int]], SExpression
    ] = None,
    symbol_to_index_fn=None,
) -> Iterator[Tuple[SExpression, List[str], PreorderMask]]:
    """
    Collects the alernate symbols that could have been selected in the tree distribution.
    """
    position = parents[-1][1]
    idxs = np.array([i for i, _ in tree_dist.distribution[parents]])
    bool_mask = mask.compute_mask(position, idxs)
    alts = tuple(int(x) for x in idxs[bool_mask])
    if replace_node_midstream is not None:
        s_exp = replace_node_midstream(s_exp, mask, position, alts)
    yield s_exp, alts, mask
    sym_idx = (
        tree_dist.symbol_to_index[s_exp.symbol]
        if symbol_to_index_fn is None
        else symbol_to_index_fn(mask, s_exp.symbol)
    )
    mask.on_entry(position, sym_idx)
    order = tree_dist.ordering.order(sym_idx, len(s_exp.children))
    for idx, child in zip(order, [s_exp.children[i] for i in order]):
        new_parents = (parents + ((sym_idx, idx),))[-tree_dist.limit :]
        yield from collect_preorder_symbols_dfs(
            child,
            tree_dist,
            mask,
            new_parents,
            replace_node_midstream=replace_node_midstream,
            symbol_to_index_fn=symbol_to_index_fn,
        )
    mask.on_exit(position, sym_idx)


def annotate_with_alternate_symbols(
    s_exp: SExpression,
    tree_dist: TreeDistribution,
    summary_fn=lambda chosen, alts: f"{chosen}/{','.join(sorted(alts))}",
) -> SExpression:
    """
    Annotates the S-Expression with the alternate symbols that could have been
    selected in the tree distribution.
    """
    preorder_symbols = [
        (node, alts) for node, alts, _ in collect_preorder_symbols(s_exp, tree_dist)
    ]
    assert len(preorder_symbols) == len({id(node) for node, _ in preorder_symbols})
    node_id_to_alts = {
        id(node): tuple(tree_dist.symbols[alt][0] for alt in alts)
        for node, alts in preorder_symbols
    }

    def replace(s):
        symbol = summary_fn(s.symbol, node_id_to_alts[id(s)])
        children = [replace(c) for c in s.children]
        return SExpression(symbol, children)

    return replace(s_exp)



from dataclasses import dataclass
from typing import List

from neurosym.dsl.abstraction import AbstractionIndexParameter, AbstractionParameter
from neurosym.dsl.dsl import DSL
from neurosym.dsl.production import LambdaProduction, VariableProduction
from neurosym.programs.s_expression import InitializedSExpression
from neurosym.types.type_signature import LambdaTypeSignature


@dataclass
class LambdaFunction:
    dsl: DSL
    body: InitializedSExpression
    typ: LambdaTypeSignature

    @classmethod
    def of(cls, dsl, body, typ):
        return cls(dsl, body, typ)

    def __call__(self, *args):
        assert len(args) == self.typ.function_arity()
        body = replace_variables_with_args(self.dsl, self.body)
        # Reverse the arguments because we want to inject them from the right.
        body = inject(self.dsl, body, args[::-1])
        return self.dsl.compute(body)


def replace_variables_with_args(dsl: DSL, expr: InitializedSExpression):
    """
    Replace variable nodes with AbstractionIndexParameter nodes.
    """
    if isinstance(expr, AbstractionParameter):
        return expr
    if isinstance(expr, AbstractionIndexParameter):
        return expr
    assert isinstance(expr, InitializedSExpression)
    prod = dsl.get_production(expr.symbol)
    if isinstance(prod, VariableProduction):
        return AbstractionIndexParameter(prod.index_in_env)
    return InitializedSExpression(
        expr.symbol,
        [replace_variables_with_args(dsl, child) for child in expr.children],
        expr.state,
    )


def inject(dsl: DSL, expr: InitializedSExpression, args: List[object], frame_shift=0):
    """
    Inject the arguments into the expression, and frame shift the indices.
    """
    if isinstance(expr, AbstractionParameter):
        return expr
    if isinstance(expr, AbstractionIndexParameter):
        assert expr.index - frame_shift < len(args)
        if expr.index - frame_shift >= 0:
            return AbstractionParameter(args[expr.index - frame_shift])
        return expr
    prod = dsl.get_production(expr.symbol)
    children = expr.children
    if isinstance(prod, LambdaProduction):
        frame_shift += prod.type_signature().function_arity()
    return InitializedSExpression(
        expr.symbol,
        [inject(dsl, child, args, frame_shift=frame_shift) for child in children],
        expr.state,
    )



import ast
import base64
from typing import Callable, Dict, List, Union

from frozendict import frozendict
from increase_recursionlimit import increase_recursionlimit

from neurosym.programs.s_expression import SExpression
from neurosym.programs.s_expression_render import parse_s_expression
from neurosym.python_dsl import python_ast_tools
from neurosym.python_dsl.names import PYTHON_DSL_SEPARATOR

from .python_ast import (
    LeafAST,
    ListAST,
    NodeAST,
    PythonAST,
    SequenceAST,
    SliceElementAST,
    SpliceAST,
    StarrableElementAST,
)
from .symbol import PythonSymbol


def s_exp_leaf_to_value(x):
    """
    Returns (True, a python representation of the leaf) if it is a leaf,
        or (False, None) otherwise.
    """
    sym_x = PythonSymbol.parse(x)
    if sym_x is not None:
        return True, sym_x
    if x == "Ellipsis":
        return True, Ellipsis
    if x in {"True", "False", "None"}:
        return True, ast.literal_eval(x)
    if x.startswith("i"):
        return True, int(x[1:])
    if x.startswith("f"):
        return True, float(x[1:])
    if x.startswith("j"):
        return True, complex(x[1:])
    if x.startswith("s_"):
        return True, x[2:]
    if x.startswith("s-"):
        return True, "".join(
            chr(x)
            for x in ast.literal_eval(
                base64.b64decode(x[2:].encode("utf-8")).decode("ascii")
            )
        )
    if x.startswith("b"):
        return True, base64.b64decode(x[1:].encode("utf-8"))

    return False, None


def handle_leaf(
    x: str,
    node_hooks: Dict[str, Callable[[str, List[PythonAST]], PythonAST]],
):
    for hook_prefix, hook in node_hooks.items():
        if x.startswith(hook_prefix):
            return hook(x, [])

    is_leaf, leaf = s_exp_leaf_to_value(x)
    if is_leaf:
        return LeafAST(leaf)
    typ = getattr(ast, x)
    assert not python_ast_tools.fields_for_node(typ), typ
    return NodeAST(typ, [])


def s_exp_to_parsed_ast(
    x: SExpression,
    node_hooks: Dict[str, Callable[[str, List[PythonAST]], PythonAST]],
) -> PythonAST:
    """
    Convert an s-expression (as pairs) to a parsed AST
    """
    if x == "nil":
        return ListAST([])
    if isinstance(x, str):
        return handle_leaf(x, node_hooks)
    assert isinstance(x, SExpression), str((type(x), x))
    tag, args = x.symbol, x.children
    # remove any type information
    tag = tag.split(PYTHON_DSL_SEPARATOR)[0]
    if tag.startswith("const-"):
        assert len(args) == 0
        is_leaf, leaf = s_exp_leaf_to_value(tag[len("const-") :])
        assert is_leaf
        return LeafAST(leaf)
    assert isinstance(tag, str), str(tag)
    args = [s_exp_to_parsed_ast(x, node_hooks) for x in args]
    if tag in {"/seq", "/subseq", "/choiceseq"}:
        return SequenceAST(tag, args)
    if tag in {"/splice"}:
        [arg] = args
        return SpliceAST(arg)
    if tag.startswith("_slice"):
        assert len(args) == 1
        return SliceElementAST(args[0])
    if tag.startswith("_starred"):
        assert len(args) == 1
        return StarrableElementAST(args[0])
    if tag in {"list"}:
        return ListAST(args)
    for hook_prefix, hook in node_hooks.items():
        if tag.startswith(hook_prefix):
            return hook(tag, args)
    return NodeAST(getattr(ast, tag), args)


def s_exp_to_python_ast(
    code: Union[str, SExpression],
    node_hooks: Dict[str, Callable[[str, List[PythonAST]], PythonAST]] = frozendict(),
) -> PythonAST:
    """
    Converts an s expression to a PythonAST object. If the code is a string,
        it is first parsed into an s-expression.
    """
    with increase_recursionlimit():
        if isinstance(code, str):
            code = parse_s_expression(code)
        code = s_exp_to_parsed_ast(code, node_hooks)
        return code



"""
Check that all neural modules have desired input-output signatures.
"""
import unittest

import torch

from neurosym.examples import near


class TestNeuralModels(unittest.TestCase):
    input_dim = 10
    output_dim = 4
    hidden_dim = 20
    traj_len = 3
    bs = 2

    def test_mlp(self):
        cfg = near.MLPConfig("mlp", self.input_dim, self.hidden_dim, self.output_dim)
        mlp = near.MLP(cfg)
        x = torch.randn(self.bs, self.traj_len, self.input_dim)
        self.assertEqual(mlp(x).shape, (self.bs, self.traj_len, self.output_dim))

    def test_seq2class_rnn(self):
        cfg = near.RNNConfig("rnn", self.input_dim, self.hidden_dim, self.output_dim)
        seq2class_rnn = near.Seq2ClassRNN(cfg)
        x = torch.randn(self.bs, self.traj_len, self.input_dim)
        self.assertEqual(seq2class_rnn(x).shape, (self.bs, self.output_dim))

    def test_seq2seq_rnn(self):
        cfg = near.RNNConfig("rnn", self.input_dim, self.hidden_dim, self.output_dim)
        seq2seq_rnn = near.Seq2SeqRNN(cfg)
        x = torch.randn(self.bs, self.traj_len, self.input_dim)
        self.assertEqual(
            seq2seq_rnn(x).shape, (self.bs, self.traj_len, self.output_dim)
        )
